# plot a pdf, cdf, and survival function for a weibull distribution (in three subplots)
from scipy.stats import norm, weibull_min
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

# set seed
np.random.seed(111)

# create a range of x values
x = np.linspace(0, 6, 100)

# create a weibull distribution with shape 1.5 and scale 2
dist = weibull_min(1.5, scale=1.3)

# create a figure and a set of subplots
fig, ax = plt.subplots(nrows=1, ncols=3, figsize=(9, 3))

# plot the probability density function
# ax[0].spines['left'].set_position('center')
ax[0].spines['right'].set_color('none')
ax[0].spines['top'].set_color('none')
ax[0].plot(x, dist.pdf(x), zorder = 10)
ax[0].set_ylabel('Probability Density Function')
ax[0].set_xlabel('time')
ax[0].set_yticks([0, 0.2, 0.4, 0.6, 0.8, 1], minor=False)
ax[0].set_xlim(0.0, 6.0)
ax[0].set_ylim(0.0, 1.0)
ax[0].grid(axis='y')

# # plot the cumulative distribution function
# # ax[1].spines['left'].set_position('center')
# ax[1].spines['right'].set_color('none')
# ax[1].spines['top'].set_color('none')
# ax[1].plot(x, dist.cdf(x), zorder = 10)
# ax[1].set_title('Cumulative Density Function')
# ax[1].set_yticks([0, 0.25, 0.5, 0.75, 1], minor=False)
# ax[1].set_ylim(0.0, 1.0)
# ax[1].set_xlabel('time')
# ax[1].grid(axis='y')

# plot the survival function
# ax[2].spines['left'].set_position('center')
ax[1].spines['right'].set_color('none')
ax[1].spines['top'].set_color('none')
ax[1].plot(x, dist.sf(x),  zorder = 10)
ax[1].set_ylabel('Survival Function')
ax[1].set_yticks([0, 0.2, 0.4, 0.6, 0.8, 1], minor=False)
ax[1].set_xlabel('time')
ax[1].set_xlim(0.0, 6.0)
ax[1].set_ylim(0.0, 1.0)
ax[1].grid(axis='y')


# plot the cumulative hazard function
ax[2].spines['right'].set_color('none')
ax[2].spines['top'].set_color('none')
ax[2].plot(x, -dist.logsf(x),  zorder = 10)
ax[2].set_ylabel('Cumulative Hazard Function')
# ax[1].set_yticks([0, 0.2, 0.4, 0.6, 0.8, 1], minor=False)
ax[2].set_xlabel('time')
ax[2].set_xlim(0.0, 6.0)
ax[2].set_ylim(0.0, 10.0)
ax[2].grid(axis='y')

colors = sns.color_palette('colorblind', 8)

# draw 100 samples from the weibull distribution
samples_40 = dist.rvs(100)

dot_size = 8
ax[0].scatter(samples_40, dist.pdf(samples_40), color=colors[7], marker='o', s=dot_size, zorder=8)
# ax[1].scatter(samples_40, dist.cdf(samples_40), color=colors[7], marker='o', s=dot_size)
ax[1].scatter(samples_40, dist.sf(samples_40), color=colors[7], marker='o', s=dot_size, zorder=8)
ax[2].scatter(samples_40, -dist.logsf(samples_40), color=colors[7], marker='o', s=dot_size, zorder=8)

# plot the sideway histogram for the
ax0 = ax[0].twiny()
# ax0.get_xaxis().set_visible(False)
# ax0.set_xticks([])

ax1 = ax[1].twiny()
ax1.hist(dist.sf(samples_40), bins=[0, 0.2, 0.4, 0.6, 0.8, 1], density=False, alpha=0.3, color='gray', orientation='horizontal', zorder=5)
ax1.yaxis.set_ticks_position('left')
# ax1.set_yticks([0, 0.2, 0.4, 0.6, 0.8, 1], minor=False)
ax1.set_ylim(0.0, 1.0)
# ax1.set_xlim(0, 1)
ax1.xaxis.set_ticks_position('top')
ax1.set_xlabel('count')

ax2 = ax[2].twiny()
ax2.hist(-dist.logsf(samples_40), bins=[0, 2, 4, 6, 8, 10], density=False, alpha=0.3, color='gray', orientation='horizontal', zorder=5)
ax2.yaxis.set_ticks_position('left')
ax2.set_ylim(0.0, 10.0)
ax2.xaxis.set_ticks_position('top')
ax2.set_xlabel('count')
# ax2.set_xlim(0, 1)


plt.tight_layout()
plt.savefig('weibull_distribution_40.png', dpi=400)

surv_probs = dist.sf(samples_40)
print("Above 75%:", len(surv_probs[surv_probs > 0.75]))
print("50%-75%:", len(surv_probs[(surv_probs > 0.5) & (surv_probs <= 0.75)]))
print("25%-50%:", len(surv_probs[(surv_probs > 0.25) & (surv_probs <= 0.5)]))
print("Below 25%:", len(surv_probs[surv_probs <= 0.25]))





import lifelines
from lifelines import CoxPHFitter
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from scipy.integrate import cumulative_trapezoid
import math
import seaborn as sns
from sympy.printing.pretty.pretty_symbology import line_width

from SurvivalEVAL import LifelinesEvaluator


plt.style.use('seaborn-v0_8-paper')
plt.rcParams["figure.figsize"] = (8, 3.5)
plt.rc('axes', titlesize=9.5)     # fontsize of the axes title
plt.rc('axes', labelsize=9.5)    # fontsize of the x and y labels
# plt.rc('xtick', labelsize=10)    # fontsize of the tick labels
# plt.rc('ytick', labelsize=10)    # fontsize of the tick labels
plt.rc('legend', fontsize=9.5)    # legend fontsize
plt.rc('figure', titlesize=9.5)  # fontsize of the figure title

plt.rcParams['font.size'] = 12

#%% 4 patients ISD
color_list = sns.color_palette("deep", 6)
# custom_params = {"axes.spines.right": False, "axes.spines.top": False,}
# sns.set_theme(style="white")

height_unit = 3
width_unit = 3.5


def weibull_pdf(x, scale, shape):
    return (shape / scale) * (x / scale)**(shape - 1) * np.exp(-(x / scale)**shape)


def gamma_pdf(x, scale, shape):
    return (x ** (shape - 1) * np.exp(-x / scale)) / (math.gamma(shape) * scale ** shape)


def lognormal_pdf(x, mu, sigma):
    return (1 / (x * sigma * np.sqrt(2 * np.pi))) * np.exp(-(np.log(x) - mu) ** 2 / (2 * sigma ** 2))


e_1 = 1.5
e_2 = 1.8
e_3 = 0.8
e_4 = 2.15
end_time = 3
N = 100000
s1 = 0.5 * np.random.weibull(a=1, size=int(0.4*N))   # generate your data sample with N elements
s2 = 2.8 * np.random.weibull(a=7, size=int(0.6*N))
sample_all = np.append(s1, s2)
kde = gaussian_kde(sample_all)
# these are the values over which your kernel will be evaluated
dist_space = np.arange(0, end_time, 0.001)
p_1 = kde(dist_space)
# plt.plot( dist_space, pdf )
cdf_1 = cumulative_trapezoid(p_1, dist_space, initial=0)
death_idx_1 = int(e_1/0.001)
# ab_1 = AnnotationBbox(death_imagebox, (e_1, 1 - cdf_1[death_idx_1]), frameon = False)

p_2 = weibull_pdf(dist_space, 1.1, 2)
cdf_2 = cumulative_trapezoid(p_2, dist_space, initial=0)
death_idx_2 = int(e_2/0.001)
p_3 = gamma_pdf(dist_space, 0.6, 3)
cdf_3 = cumulative_trapezoid(p_3, dist_space, initial=0)
death_idx_3 = int(e_3/0.001)
p_4 = lognormal_pdf(dist_space, 0.7, 0.2)
p_4[0] = 0
cdf_4 = cumulative_trapezoid(p_4, dist_space, initial=0)
death_idx_4 = int(e_4/0.001)

# D-calibration figures
fig, ax = plt.subplots(nrows=1, ncols=3, figsize=(width_unit*2.5, height_unit))

# d-calibration
ax[0].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0], linewidth=2, clip_on=False, zorder=10)
ax[0].plot(dist_space, 1 - cdf_2, label='Patient B', c=color_list[1], linewidth=2, clip_on=False, zorder=10)
ax[0].plot(dist_space, 1 - cdf_3, label='Patient C', c=color_list[2], linewidth=2, clip_on=False, zorder=10)
ax[0].plot(dist_space, 1 - cdf_4, label='Patient D', c=color_list[3], linewidth=2, clip_on=False, zorder=10)
ax[0].plot(e_1, 1 - cdf_1[death_idx_1], marker='^', markersize=10, c=color_list[0])
ax[0].plot(e_2, 1 - cdf_2[death_idx_2], marker='v', markersize=10, c=color_list[1])
ax[0].plot(e_3, 1 - cdf_3[death_idx_3], marker='v', markersize=10, c=color_list[2])
ax[0].plot(e_4, 1 - cdf_4[death_idx_4], marker='v', markersize=10, c=color_list[3])

ax[0].grid(False)
# ax[1, 2].spines.set_color('black')
ax[0].set_ylim([0, 1])
ax[0].set_xlim([0, end_time])
ax[0].set_yticks([0, 0.25, 0.5, 0.75, 1])
ax[0].set_xlabel('Days')
ax[0].set_ylabel('Survival Probability')
ax[0].set_title('(a) Prediction')

probs_at_event = np.array([1 - cdf_1[death_idx_1],
                           1 - cdf_2[death_idx_2],
                           1 - cdf_3[death_idx_3],
                           1 - cdf_4[death_idx_4]])
probs ={
    'Null': np.array([0, 0, 0, 0]),
    'B': np.array([1, 0, 0, 0]),
    'C': np.array([0, 0, 0, 1]),
    'D': np.array([0, 1, 0, 0]),
    'A': np.array([0.25 / (1 - cdf_1[death_idx_1]), 0.25 / (1 - cdf_1[death_idx_1]),
                   (1 - cdf_1[death_idx_1] - 0.5) / (1 - cdf_1[death_idx_1]), 0]),
}
left = np.zeros_like(probs_at_event)

for i, (key, value) in enumerate(probs.items()):
    if key == 'A':
        i = 0
    ax[1].barh([0, 0.25, 0.5, 0.75], value, height=0.25, color=color_list[i], label=key,
              align='edge', edgecolor='black', left=left, linewidth=1)
    left += value
# ax[1].hist(probs_at_event, bins=[0, 0.25, 0.5, 0.75, 1], color=color_list[5], rwidth=1, orientation='horizontal')
ax[1].set_ylim([0, 1])
ax[1].axvline(1, ls='dashed', c='grey', label='Perfect Cal.')
ax[1].set_xlabel('Counts')
# ax[1].set_ylabel('Survival Probability')
ax[1].grid(False)
ax[1].set_xlim([0, 2.5])
ax[1].set_xticks([0, 1, 2])
ax[1].set_xticklabels([0, 1, 2])
ax[1].set_yticks([0, 0.25, 0.5, 0.75, 1])
# ax[1].legend(loc='lower right')
ax[1].set_title('(b) D-calibration Histogram')

# hist = np.histogram(probs_at_event, bins=[0, 0.25, 0.5, 0.75, 1])
probs = left / 4
cum_probs = np.cumsum(probs)
cum_probs = np.insert(cum_probs, 0, 0)
ax[2].plot([0, 1], [0, 1], ls='dashed', c='grey', label='Ideal')
ax[2].plot([0, 0.25, 0.5, 0.75, 1], cum_probs, marker='o', markersize=8, c='skyblue', linewidth=2,
              clip_on=False, zorder=10)
ax[2].set_ylim([0, 1])
ax[2].set_xlim([0.0, 1.0])
ax[2].set_xticks([0, 0.25, 0.5, 0.75, 1])
ax[2].set_yticks([0, 0.25, 0.5, 0.75, 1])
ax[2].set_xlabel('Predicted Survival Probability')
ax[2].set_ylabel('Observed Survival Probability')
ax[2].grid(False)
ax[2].legend(loc='lower right')
ax[2].set_title('(c) D-calibration P-P Plot')


plt.tight_layout()
plt.savefig(f'd-cal_censor.png', dpi=400, transparent=True)

#%% real data example

# ==============================================================================
# 1. Load the Data
# ==============================================================================
data = lifelines.datasets.load_gbsg2()
# preprocessing
data.rename(columns={'cens': 'event'}, inplace=True)
data['horTh'] = data['horTh'].map({'no': 0, 'yes': 1})
data['menostat'] = data['menostat'].map({'Pre': 0, 'Post': 1})
data['tgrade'] = data['tgrade'].map({'I': 1, 'II': 2, 'III': 3})
# randomly divide the data into training and validation sets
df_train = data.sample(frac=0.7, random_state=42)  # 70% for training
df_train = df_train.reset_index(drop=True)
df_test = data.drop(df_train.index)  # remaining 30% for testing
df_test = df_test.reset_index(drop=True)
x_test = df_test.drop(columns=['time', 'event']).values

# ==============================================================================
# 2. Fit a Cox Proportional Hazards Model
# ==============================================================================
cph = CoxPHFitter()
cph.fit(df_train, duration_col='time', event_col='event')
year = 3  # Set the target year for calibration
# Define time points for 1, 2, 3, 4, 5, 6 years (in days)
times = 365 * year * np.arange(1, 7)

# Predict the survival probability at the given time points
survs_cox = cph.predict_survival_function(x_test)

# pass to evaluator
evler = LifelinesEvaluator(survs_cox, df_test['time'].values, df_test['event'].values)

# ==============================================================================
# residuals
p_value, combined_binning = evler.d_calibration(10)
predict_probs = combined_binning / combined_binning.sum()  # normalize the counts to get probabilities

# probs_at_event = combined_binning / combined_binning.sum()
fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(7, 3))

# sideway histogram using combined binning
bins = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]
#combined_binning has a size of 10, each one indicate the counts of the bin
ax[0].barh(bins[:-1], predict_probs, height=0.1, color='#CDE7F0', align='edge', edgecolor='black', label='Counts',
           linewidth=1)
# ax[0].hist(predict_probs, bins=bins, color=color_list[5], rwidth=1, orientation='horizontal', density=True)
# ax[0].set_ylim([-0.05, 1.05])
ax[0].axvline(0.1, ls='dashed', c='grey', label='Perfect Cal.')
ax[0].set_xlabel('Proportion')
ax[0].set_yticks([0, 0.2, 0.4, 0.6, 0.8, 1])
# ax[0].set_ylabel('Survival Probability')
ax[0].grid(False)
# ax[0].set_xlim([0, 0.15])
ax[0].set_xticks([0, 0.05, 0.1, 0.15])
ax[0].set_xticklabels(['0%', '5%', '10%', '15%'])
# ax[0].legend(loc='lower right')
ax[0].set_ylabel('Predicted Survival Probability')
ax[0].set_title('(a) D-calibration Histogram')

# p-p calibration curve

cum_probs = np.cumsum(predict_probs)
cum_probs = np.insert(cum_probs, 0, 0)
ax[1].plot([0, 1], [0, 1], ls='dashed', c='grey', label='Ideal')
ax[1].plot(bins, cum_probs, marker='o', markersize=8, c='skyblue', linewidth=2,
              clip_on=False, zorder=10)
ax[1].text(0.05, 0.95, f'p-value: {p_value:.3f}', fontsize=9.5, color='black', ha='left', va='top')
# ax[1].set_ylim([-0.05, 1.05])
# ax[1].set_xlim([-0.0, 1.0])
ax[1].set_xticks([0, 0.2, 0.4, 0.6, 0.8, 1])
ax[1].set_yticks([0, 0.2, 0.4, 0.6, 0.8, 1])
ax[1].set_xlabel('Predicted Survival Probability')
ax[1].set_ylabel('Observed Survival Probability')
ax[1].grid(False)
ax[1].legend(loc='lower right')
ax[1].set_title('(b) D-calibration P-P Plot')

plt.tight_layout()
plt.savefig('d-cal_cox.png', dpi=400)



from lifelines import CoxPHFitter
from lifelines.datasets import load_rossi

import numpy as np
from SurvivalEVAL.Evaluator import LifelinesEvaluator

np.random.seed(42)
# Load the data
rossi = load_rossi()
rossi = rossi.sample(frac=1.0)

# Split train/test set
train = rossi.iloc[:300, :]
test = rossi.iloc[300:, :]
train_event_times = train.week.values
train_event_indicators = train.arrest.values
test_event_times = test.week.values
test_event_indicators = test.arrest.values

# Fit the model
cph = CoxPHFitter()
cph.fit(train, duration_col='week', event_col='arrest')

survival_curves = cph.predict_survival_function(test)

eval = LifelinesEvaluator(survival_curves, test_event_times, test_event_indicators,
                          train_event_times, train_event_indicators, predict_time_method="Median")

# Make the evaluation
cindex, _, _ = eval.concordance(method="Margin")

bs = eval.brier_score(IPCW_weighted=True)
# The largest event time is 52. So we use 53 time points (0, 1, ..., 52) to calculate the IBS
ibs = eval.integrated_brier_score(num_points=53, IPCW_weighted=False, draw_figure=True)

l1 = eval.mae(method="Margin")

one_cal = eval.one_calibration(target_time=25)

d_cal = eval.d_calibration()



import numpy as np
from typing import Optional

from SurvivalEVAL.NonparametricEstimator.SingleEvent import KaplanMeierArea


def concordance(
        predicted_times: np.ndarray,
        event_times: np.ndarray,
        event_indicators: np.ndarray,
        train_event_times: Optional[np.ndarray] = None,
        train_event_indicators: Optional[np.ndarray] = None,
        method: str = "Harrell",
        ties: str = "Risk"
) -> (float, float, int):
    """
    Calculate the concordance index between the predicted survival times and the true survival times.

    Parameters
    ----------
    predicted_times: np.ndarray, shape = (n_samples,)
        The predicted survival times.
    event_times: np.ndarray, shape = (n_samples,)
        The true survival times.
    event_indicators: np.ndarray, shape = (n_samples,)
        The event indicators of the true survival times.
    train_event_times: np.ndarray, shape = (n_train_samples,)
        The true survival times of the training set.
    train_event_indicators: np.ndarray, shape = (n_train_samples,)
        The event indicators of the true survival times of the training set.
    method: str, optional (default="Harrell")
        A string indicating the method for constructing the pairs of samples.
        "Harrell": the pairs are constructed by comparing the predicted survival time of each sample with the
        event time of all other samples. The pairs are only constructed between samples with comparable
        event times. For example, if i-th sample has a censor time of 10, then the pairs are constructed by
        comparing the predicted survival time of sample i with the event time of all samples with event
        time of 10 or less.
        "Margin": the pairs are constructed between all samples. A best-guess time for the censored samples
        will be calculated and used to construct the pairs.
    ties: str, optional (default="Risk")
        A string indicating the way ties should be handled.
        Options: "None" (default), "Time", "Risk", or "All"
        "None" will throw out all ties in true survival time and all ties in predict survival times (risk scores).
        "Time" includes ties in true survival time but removes ties in predict survival times (risk scores).
        "Risk" includes ties in predict survival times (risk scores) but not in true survival time.
        "All" includes all ties.
        Note the concordance calculation is given by
        (Concordant Pairs + (Number of Ties/2))/(Concordant Pairs + Discordant Pairs + Number of Ties).

    Returns
    -------
    c_index: float
        The concordance index.
    concordant_pairs: float
        The number of concordant pairs.
    total_pairs: int
        The total number of comparable pairs.
    """
    # the scikit-survival concordance function only takes risk scores to calculate.
    # So at first we should transfer the predicted time -> risk score.
    # The risk score should be higher for subjects that live shorter (i.e. lower average survival time).

    event_indicators = event_indicators.astype(bool)

    assert len(predicted_times) == len(event_times) == len(event_indicators), \
        "The lengths of the predicted times and labels must be the same."

    if method == "Harrell":
        risks = -1 * predicted_times
        partial_weights = None
        bg_event_times = None
    elif method == "Margin":
        if train_event_times is None or train_event_indicators is None:
            error = "If 'Margin' is chosen, training set information must be provided."
            raise ValueError(error)

        train_event_indicators = train_event_indicators.astype(bool)

        km_model = KaplanMeierArea(train_event_times, train_event_indicators)
        km_linear_zero = -1 / ((1 - min(km_model.survival_probabilities))/(0 - max(km_model.survival_times)))
        if np.isinf(km_linear_zero):
            km_linear_zero = max(km_model.survival_times)
        predicted_times = np.clip(predicted_times, a_max=km_linear_zero, a_min=None)
        risks = -1 * predicted_times

        censor_times = event_times[~event_indicators]
        partial_weights = np.ones_like(event_indicators, dtype=float)
        partial_weights[~event_indicators] = 1 - km_model.predict(censor_times)

        best_guesses = km_model.best_guess(censor_times)
        best_guesses[censor_times > km_linear_zero] = censor_times[censor_times > km_linear_zero]

        bg_event_times = np.copy(event_times)
        bg_event_times[~event_indicators] = best_guesses
    else:
        raise TypeError("Method for calculating concordance is unrecognized.")
    # risk_ties means predicted times are the same while true times are different.
    # time_ties means true times are the same while predicted times are different.
    # c_index, concordant_pairs, discordant_pairs, risk_ties, time_ties = metrics.concordance_index_censored(
    #     event_indicators, event_times, estimate=risk)
    c_index, concordant_pairs, discordant_pairs, risk_ties, time_ties = _estimate_concordance_index(
        event_indicators, event_times, estimate=risks, bg_event_time=bg_event_times, partial_weights=partial_weights)
    if ties == "None":
        total_pairs = concordant_pairs + discordant_pairs
        c_index = concordant_pairs / total_pairs
    elif ties == "Time":
        total_pairs = concordant_pairs + discordant_pairs + time_ties
        concordant_pairs = concordant_pairs + 0.5 * time_ties
        c_index = concordant_pairs / total_pairs
    elif ties == "Risk":
        # This should be the same as original outputted c_index from above
        total_pairs = concordant_pairs + discordant_pairs + risk_ties
        concordant_pairs = concordant_pairs + 0.5 * risk_ties
        c_index = concordant_pairs / total_pairs
    elif ties == "All":
        total_pairs = concordant_pairs + discordant_pairs + risk_ties + time_ties
        concordant_pairs = concordant_pairs + 0.5 * (risk_ties + time_ties)
        c_index = concordant_pairs / total_pairs
    else:
        error = "Please enter one of 'None', 'Time', 'Risk', or 'All' for handling ties for concordance."
        raise TypeError(error)

    return c_index, concordant_pairs, total_pairs


def _estimate_concordance_index(
        event_indicator: np.ndarray,
        event_time: np.ndarray,
        estimate: np.ndarray,
        bg_event_time: np.ndarray = None,
        partial_weights: np.ndarray = None,
        tied_tol: float = 1e-8
) -> (float, float, float, float, float):
    """
    Estimate the concordance index.
    This backbone of this function is borrowed from scikit-survival:
    https://github.com/sebp/scikit-survival/blob/4e664d8e4fe5e5b55006e3913f2bbabcf2455496/sksurv/metrics.py#L85-L118
    In here, we make modifications to the original function to allow for partial weights and best-guess times (margin times).

    All functions in scikit-survival are licensed under the GPLv3 License:
    https://github.com/sebp/scikit-survival/blob/master/COPYING

    Parameters
    ----------
    event_indicator: np.ndarray, shape = (n_samples,)
        The event indicators of the true survival times.
    event_time: np.ndarray, shape = (n_samples,)
        The true survival times.
    estimate: np.ndarray, shape = (n_samples,)
        The estimated risk scores. A higher score should correspond to a higher risk.
    bg_event_time: np.ndarray, shape = (n_samples,), optional (default=None)
        The best-guess event times. For uncensored samples, this should be the same as the true event times.
        For censored samples, this should be the best-guess time (margin time) for the censored samples.
    partial_weights: np.ndarray, shape = (n_samples,), optional (default=None)
        The partial weights for the censored samples.
    tied_tol: float, optional (default=1e-8)
        The tolerance for considering two times as tied.

    Returns
    -------
    c_index: float
        The concordance index.
    concordant: float
        The number of concordant pairs.
    discordant: float
        The number of discordant pairs.
    tied_risk: float
        The number of tied risk scores.
    tied_time: float
        The number of tied times.
    -------
    """
    order = np.argsort(event_time, kind="stable")

    comparable, tied_time, weight = _get_comparable(event_indicator, event_time, order)

    if partial_weights is not None:
        event_indicator = np.ones_like(event_indicator)
        comparable_2, tied_time, weight = _get_comparable(event_indicator, bg_event_time, order, partial_weights)
        for ind, mask in comparable.items():
            weight[ind][mask] = 1
        comparable = comparable_2

    if len(comparable) == 0:
        raise ValueError("Data has no comparable pairs, cannot estimate concordance index.")

    concordant = 0
    discordant = 0
    tied_risk = 0
    numerator = 0.0
    denominator = 0.0
    for ind, mask in comparable.items():
        est_i = estimate[order[ind]]
        event_i = event_indicator[order[ind]]
        # w_i = partial_weights[order[ind]] # change this
        w_i = weight[ind]
        weight_i = w_i[order[mask]]

        est = estimate[order[mask]]

        assert event_i, 'got censored sample at index %d, but expected uncensored' % order[ind]

        ties = np.absolute(est - est_i) <= tied_tol
        # n_ties = ties.sum()
        n_ties = np.dot(weight_i, ties.T)
        # an event should have a higher score
        con = est < est_i
        # n_con = con[~ties].sum()
        con[ties] = False
        n_con = np.dot(weight_i, con.T)

        # numerator += w_i * n_con + 0.5 * w_i * n_ties
        # denominator += w_i * mask.sum()
        numerator += n_con + 0.5 * n_ties
        denominator += np.dot(w_i, mask.T)

        tied_risk += n_ties
        concordant += n_con
        # discordant += est.size - n_con - n_ties
        discordant += np.dot(w_i, mask.T) - n_con - n_ties

    c_index = numerator / denominator
    return c_index, concordant, discordant, tied_risk, tied_time


def _get_comparable(
        event_indicator: np.ndarray,
        event_time: np.ndarray,
        order: np.ndarray,
        partial_weights: np.ndarray = None
) -> (dict, int, dict):
    """
    Given the labels of the survival outcomes, get the comparable pairs.

    This backbone of this function is borrowed from scikit-survival:
    https://github.com/sebp/scikit-survival/blob/4e664d8e4fe5e5b55006e3913f2bbabcf2455496/sksurv/metrics.py#L57-L81
    In here, we make modifications to the original function to calculates the weights for each pair.

    All functions in scikit-survival are licensed under the GPLv3 License:
    https://github.com/sebp/scikit-survival/blob/master/COPYING

    Parameters
    ----------
    event_indicator: np.ndarray, shape = (n_samples,)
        The event indicators of the true survival times.
    event_time: np.ndarray, shape = (n_samples,)
        The true survival times.
    order: np.ndarray, shape = (n_samples,)
        The indices that would sort the event times.
    partial_weights: np.ndarray, shape = (n_samples,), optional (default=None)
        The partial weights for the censored samples.

    Returns
    -------
    comparable: dict
        A dictionary where the keys are the indices of the samples with events, and the values are boolean masks
        indicating which samples are comparable to the key sample.
    tied_time: int
        The number of tied times.
    weight: dict
        A dictionary where the keys are the indices of the samples with events, and the values are the weights
        for each comparable sample.
    """

    if partial_weights is None:
        partial_weights = np.ones_like(event_indicator, dtype=float)
    n_samples = len(event_time)
    tied_time = 0
    comparable = {}
    weight = {}

    i = 0
    while i < n_samples - 1:
        time_i = event_time[order[i]]
        end = i + 1
        while end < n_samples and event_time[order[end]] == time_i:
            end += 1

        # check for tied event times
        event_at_same_time = event_indicator[order[i:end]]
        censored_at_same_time = ~event_at_same_time

        for j in range(i, end):
            if event_indicator[order[j]]:
                mask = np.zeros(n_samples, dtype=bool)
                mask[end:] = True
                # an event is comparable to censored samples at same time point
                mask[i:end] = censored_at_same_time
                comparable[j] = mask
                tied_time += censored_at_same_time.sum()
                weight[j] = partial_weights[order] * partial_weights[order[j]]
        i = end

    return comparable, tied_time, weight



import numpy as np
from matplotlib import pyplot as plt
from scipy.stats import chisquare
from scipy.integrate import trapezoid

from SurvivalEVAL.Evaluations.util import interpolated_survival_curve
from SurvivalEVAL.NonparametricEstimator.SingleEvent import KaplanMeier, NelsonAalen


def d_calibration(
        pred_probs: np.ndarray,
        event_indicators: np.ndarray,
        num_bins: int = 10
) -> (float, np.ndarray):
    """
    Calculate the D-Calibration score.
    Parameters
    ----------
    pred_probs: np.ndarray
        The predicted survival probabilities at individual's event/censor time.
    event_indicators: np.ndarray
        The event indicators.
    num_bins: int
        The number of bins to use for the D-Calibration score.

    Returns
    -------
    pvalue: float
        The p-value of the D-Calibration test.
    combine_hist: np.ndarray
        The binning histogram of the D-Calibration test.
    """
    quantile = np.linspace(1, 0, num_bins + 1)
    censor_indicators = 1 - event_indicators

    event_probs = pred_probs[event_indicators.astype(bool)]
    event_position = np.digitize(event_probs, quantile)
    event_position[event_position == 0] = 1     # class probability==1 to the first bin

    event_hist = np.zeros([num_bins])
    for i in range(len(event_position)):
        event_hist[event_position[i] - 1] += 1

    censored_probs = pred_probs[censor_indicators.astype(bool)]

    censor_hist = np.zeros([num_bins])
    if len(censored_probs) > 0:
        for i in range(len(censored_probs)):
            partial_binning = create_censor_hist(censored_probs[i], num_bins)
            censor_hist += partial_binning

    combine_hist = event_hist + censor_hist
    _, pvalue = chisquare(combine_hist)
    return pvalue, combine_hist


def create_censor_hist(
        prob: float,
        num_bins: int
) -> np.ndarray:
    """    Create the binning histogram for a right censored instance.

    The bins are defined as follows:
    [b_0, b_1), [b_1, b_2), ..., [b_{num_bins-1}, b_num_bins] (note that the last bin is closed on both side).

    For censoring instance,
    b2 will be the infimum probability of the bin that contains S(c),
    for the bin of [b2, b3) which contains S(c), probability = (S(c) - b2) / S(c)
    for the rest of the bins, [b0, b1), [b1, b2), etc., probability = 1 / (B * S(c)), where B is the number of bins.
    :param prob: float

        The predicted probability at the censored time of a censoring instance.
    :param num_bins: int
        The number of bins to use for the D-Calibration score.
    :return:
    final_binning: np.ndarray
        The "split" histogram of this censored subject.
    """
    quantile = np.linspace(1, 0, num_bins + 1)
    censor_binning = np.zeros(num_bins)
    for i in range(num_bins):
        if prob == 1:
            censor_binning += 0.1
            break
        elif quantile[i] > prob >= quantile[i + 1]:
            first_bin = (prob - quantile[i + 1]) / prob if prob != 0 else 1
            rest_bins = 1 / (num_bins * prob) if prob != 0 else 0
            censor_binning[i] += first_bin
            censor_binning[i + 1:] += rest_bins
            break
    # assert len(censor_binning) == num_bins, f"censor binning should have size of {num_bins}"
    return censor_binning


def d_cal_ic(
        pred_probs_left: np.ndarray,
        pred_probs_right: np.ndarray,
        num_bins: int = 10
) -> (float, np.ndarray):
    """
    Calculate the D-Calibration score for interval censored data.
    Parameters
    ----------
    pred_probs_left: np.ndarray
        The predicted survival probabilities at individual's left event/censor time.
    pred_probs_right: np.ndarray
        The predicted survival probabilities at individual's right event/censor time.
        For right-censored instances, the right event/censor time is infinity, so the predicted probability is 0.
    num_bins: int
        The number of bins to use for the D-Calibration score.
    Returns
    -------
    pvalue: float
        The p-value of the D-Calibration test.
    binning: np.ndarray
        The binning histogram of the D-Calibration test.
    """
    assert len(pred_probs_left) == len(pred_probs_right), \
        "The length of pred_probs_left and pred_probs_right should have same length."

    assert np.all(pred_probs_left >= pred_probs_right), \
        "The left survival probabilities should be greater than or equal to the right survival probabilities."

    assert (np.all(pred_probs_left >= 0) and np.all(pred_probs_right >= 0) and
            np.all(pred_probs_left <= 1) and np.all(pred_probs_right <= 1)), \
        "The predicted probabilities should be in the range [0, 1]."

    n = len(pred_probs_left)

    binning = np.zeros([num_bins])
    for i in range(n):
        partial_binning = create_interval_c_hist(pred_probs_left[i], pred_probs_right[i], num_bins)
        binning += partial_binning

    _, pvalue = chisquare(binning)
    return pvalue, binning


def create_interval_c_hist(
        prob_left: float,
        prob_right: float,
        num_bins: int
) -> np.ndarray:
    """
    Create the binning histogram for an interval censored instance.
    For interval censored instance, we have two predicted probabilities: left and right.
    The left probability is the predicted survival probability at the left censored time,
    and the right probability is the predicted survival probability at the right censored time.

    The bins are defined as follows:
    [b_0, b_1), [b_1, b_2), ..., [b_{num_bins-1}, b_num_bins] (note that the last bin is closed on both side).

    (1) If the left and right probabilities are within the same bin, then the histogram will be 1 for that bin,
    and 0 for the rest of the bins.
    (2) If the left and right probabilities are in different bins, then the histogram will be split:
        - The bin (e.g., [b3, b4))that contains the left probability will have a probability of
        (prob_left - b3) / (prob_left - prob_right)
        - The bin (e.g., [b1, b2)) that contains the right probability will have a probability of
        (b2 - prob_right) / (prob_left - prob_right)
        - The intermediate bins (e.g., [b2, b3)) will have a probability of
        1 / (num_bins * (prob_left - prob_right))

    :param prob_left: float
        The predicted probability at the left censored time of an interval censoring instance.
    :param prob_right: float
        The predicted probability at the right censored time of an interval censoring instance.
    :param num_bins: int
        The number of bins to use for the D-Calibration score.
    :return:
    hist: np.ndarray
        The "split" histogram of this interval censored subject.
    """
    # make sure the left and right probabilities are in the range [0, 1]
    if prob_right == 0:
        # if the right probability is 0, then it is a right-censored instance,
        return create_censor_hist(prob_left, num_bins)
    else:
        if prob_left < prob_right:
            # enforce the natural order for survival probs
            prob_left, prob_right = prob_right, prob_left

        edges = np.linspace(1.0, 0.0, num_bins + 1)  # descending: 1, 1-1/K, ..., 0

        left_idx = np.digitize(prob_left, edges) - 1 if prob_left < 1.0 else 0
        right_idx = np.digitize(prob_right, edges) - 1
        hist = np.zeros(num_bins, dtype=float)
        if left_idx == right_idx:
            hist[left_idx] = 1.0
            return hist
        else:
            # if the left and right probabilities are in different bins
            first_hist = (prob_left - edges[left_idx + 1]) / (prob_left - prob_right)
            hist[left_idx] += first_hist

            last_hist = (edges[right_idx] - prob_right) / (prob_left - prob_right)
            hist[right_idx] += last_hist

            # fill the intermediate bins with equal probability
            if right_idx > left_idx + 1:
                intermediate_hist = (1.0 - first_hist - last_hist) / (right_idx - left_idx - 1)
                hist[left_idx + 1:right_idx] = intermediate_hist

            return hist


_residual_names = {
    "CoxSnell": "Cox-Snell Residuals",
    "Modified CoxSnell-v1": "Cox-Snell Residuals",
    "Modified CoxSnell-v2": "Cox-Snell Residuals",
    "Martingale": "Martingale Residuals",
    "Deviance": "Deviance Residuals"
}


def residuals(
        pred_probs: np.ndarray,
        event_indicators: np.ndarray,
        method: str = "CoxSnell",
        draw_figure: bool = False
) -> np.ndarray:
    """
    Calculate the residuals based on the predicted probabilities and event times.
    For details of the residuals, please refer to the Chapter 4 of the book:
    "Modelling survival data in medical research" By David Collett

    Parameters
    ----------
    pred_probs: np.ndarray
        The predicted survival probabilities at individual's event/censor time.
    event_indicators: np.ndarray
        The event indicators.
    method: str
        The method to calculate residuals. Options are "CoxSnell", "Modified CoxSnell-v1", "Modified CoxSnell-v2",
        "Martingale", "Deviance".
    draw_figure: bool
        Whether to visualize the Cox-Snell residuals and draw the cumulative hazard plot.
    Returns
    -------
    np.ndarray
        The calculated residuals.
    """
    cox_residuals = - np.log(pred_probs)

    if method == "CoxSnell":
        residuals = cox_residuals
    elif method == "Modified CoxSnell-v1" or method == "Modified CoxSnell-v2":
        # Compare with standard CoxSnell residuals, this method adds an 'excess residual' for censored instances.
        # The excess residual should also follow a unit exponential distribution, based on the lack of memory property.
        # There are two choices of excess residuals:
        # (1) use the mean of the unit exponential distribution, which is 1,
        # or (2) use the median of the unit exponential distribution, which is ln(2).
        excess_residual = 1 if method == "Modified CoxSnell-v1" else np.log(2)
        residuals = cox_residuals + excess_residual * (1 - event_indicators)
    elif method == "Martingale":
        residuals = event_indicators - cox_residuals
    elif method == "Deviance":
        def safe_log(x):
            return np.log(x + 1e-8)

        martingale_res = event_indicators - cox_residuals
        # get the sign of the martingale residuals
        sign = np.sign(martingale_res)
        residuals = sign * np.sqrt(-2 * (martingale_res + event_indicators * safe_log(cox_residuals)))
    else:
        raise ValueError("Unknown method {}".format(method))

    if draw_figure:
        cum_haz_empirical = NelsonAalen(cox_residuals, event_indicators)
        max_res = np.max(cox_residuals)
        fig, ax = plt.subplots(nrows=1, ncols=2, tight_layout=True, dpi=400)
        ax[0].plot(cum_haz_empirical.survival_times, cum_haz_empirical.cumulative_hazard, label='Empirical')
        ax[0].plot([0, max_res], [0, max_res], label='Ideal', linestyle='--', color='red')
        ax[0].legend()
        ax[0].set_xlabel('Cox-Snell Residuals')
        ax[0].set_ylabel('Cumlative hazard of residuals')

        # use solid scatter points for uncensored instances and hollow scatter points for censored instances
        idx = np.arange(len(residuals))
        event_indicators = event_indicators.astype(bool)
        ax[1].scatter(idx[event_indicators], residuals[event_indicators], alpha=0.5, color='red', label='Event')
        ax[1].scatter(idx[~event_indicators], residuals[~event_indicators], facecolors='none', edgecolors='red', alpha=0.5, label='Censored')
        ax[1].set_xlabel('Index')
        ax[1].set_ylabel(_residual_names[method])
        ax[1].legend()
        plt.show()
    return residuals


def km_calibration(
        average_survival_curve: np.ndarray,
        time_coordinates: np.ndarray,
        event_times: np.ndarray,
        event_indicators: np.ndarray,
        interpolation_method: str = 'Linear',
        draw_figure: bool = False
) -> float:
    """
    Calculate the KM calibration score between the average prediction curve and KM curve.
    The first version of KM calibration [1] is by visual inspection of the KM curve and the average curve.
    The second version of KM calibration [2] is by calculating the KL divergence between the KM curve and the average.
    This function actually calculates the (normalized) integrated mean squared error
    between the KM curve and the average prediction curve.

    This version has three benefits over [1] and [2]:
    1. It is a single number, so it is easier to compare.
    2. This calculation is symmetric (note that KL-divergence is not).
    3. The score is between 0 and 1, where 0 means perfect calibration and 1 means worst calibration.
        And the random prediction curve will have a score of 0.25.

    [1] Chapfuwa et al., Calibration and Uncertainty in Neural Time-to-Event Modeling， TNNLS， 2020
    [2] Yanagisawa, Proper Scoring Rules for Survival Analysis, ICML, 2023

    Parameters
    ----------
    average_survival_curve: np.ndarray
        The average survival curves.
    time_coordinates: np.ndarray
        The time coordinates of the average survival curves.
    event_times: np.ndarray
        The event time of the test data.
    event_indicators: np.ndarray
        The event indicator of the test data.\
    draw_figure: bool
        Whether to visualize the comparison of the KM curve and average curve.

    Returns
    -------
    mse: float
        The (normalized) integrated mean squared error between the KM curve and the average prediction curve.
    """
    unique_event_times = np.unique(event_times[event_indicators == 1])

    km_model = KaplanMeier(event_times, event_indicators)
    km_curve = km_model.predict(unique_event_times)

    # add 0 to the beginning of the time coordinates, and 1 to the beginning of the average_survival_curves
    unique_event_times = np.concatenate([[0], unique_event_times])
    km_curve = np.concatenate([[1], km_curve])
    if time_coordinates[0] != 0:
        time_coordinates = np.concatenate([[0], time_coordinates])
        average_survival_curve = np.concatenate([[1], average_survival_curve])

    # interpolate the average curve, so that it will have the same time coordinates as km_curve
    spline = interpolated_survival_curve(time_coordinates, average_survival_curve, interpolation_method)
    average_survival_curve = spline(unique_event_times)
    average_survival_curve = np.clip(average_survival_curve, 0, 1)

    # integrated over the joint time coordinates
    mse = trapezoid((average_survival_curve - km_curve) ** 2, unique_event_times)
    # normalize by the maximum time coordinate
    mse /= np.max(unique_event_times)

    if draw_figure:
        plt.plot(unique_event_times, average_survival_curve, label='Average Prediction Curve')
        plt.plot(unique_event_times, km_curve, label='KM Curve')
        plt.fill_between(unique_event_times, average_survival_curve, km_curve, alpha=0.2)
        score_text = r'KM-Calibration$= {:.3f}$'.format(mse)
        plt.plot([], [], ' ', label=score_text)
        plt.legend()
        plt.xlabel('Time')
        plt.ylabel('Survival Probability')
        plt.show()

    return mse


if __name__ == '__main__':
    ### test the KM calibration

    # # first we define the time coordinates
    # times = np.linspace(0, 100, 11)
    # # then we define the survival probabilities at each time coordinate
    # survival_probabilities = np.exp(-times / 100)
    #
    #
    # # randomly generate the event time and event indicator for 20 samples
    # num_samples = 20
    # true_t = np.random.randint(0, 100, num_samples)
    # true_e = np.random.randint(0, 2, num_samples)
    #
    # # make some random predictions between 0 and 1
    # pred_at_observed_times = np.random.rand(num_samples)
    #
    # # # calculate the KM calibration score
    # # km_calibration_score = km_calibration(survival_probabilities, times, true_t, true_e, draw_figure=True)
    # # print(km_calibration_score)
    # # calculate the Cox-Snell residuals
    # coxsnell_residuals = residuals(pred_at_observed_times, true_e, method="Martingale", draw_figure=True)

    # test interval censored D-calibration
    pred_prob_left = np.array([0.84, 0.65, 0.42, 0.75, 1, 1])
    pred_prob_right = np.array([0.72, 0.63, 0.03, 0, 0.24, 0])
    pvalue, binning = d_cal_ic(pred_prob_left, pred_prob_right, num_bins=10)


import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from scipy.integrate import cumulative_trapezoid
import math
import seaborn as sns
import matplotlib.patches as patches
import matplotlib.lines as mlines
import pandas as pd
import lifelines
from lifelines import CoxPHFitter, KaplanMeierFitter
from matplotlib.axis import Axis
from matplotlib.spines import Spine

from SurvivalEVAL import LifelinesEvaluator

color_list = sns.color_palette(n_colors=6)

plt.style.use('seaborn-v0_8-paper')
plt.rcParams["figure.figsize"] = (7, 3)
plt.rc('axes', titlesize=9.5)     # fontsize of the axes title
plt.rc('axes', labelsize=9.5)    # fontsize of the x and y labels
# plt.rc('xtick', labelsize=10)    # fontsize of the tick labels
# plt.rc('ytick', labelsize=10)    # fontsize of the tick labels
plt.rc('legend', fontsize=9.5)    # legend fontsize
plt.rc('figure', titlesize=9.5)  # fontsize of the figure title

plt.rcParams['font.size'] = 12


# ==============================================================================
# 1. Load the Data
# ==============================================================================
data = lifelines.datasets.load_gbsg2()
# preprocessing
data.rename(columns={'cens': 'event'}, inplace=True)
data['horTh'] = data['horTh'].map({'no': 0, 'yes': 1})
data['menostat'] = data['menostat'].map({'Pre': 0, 'Post': 1})
data['tgrade'] = data['tgrade'].map({'I': 1, 'II': 2, 'III': 3})
# randomly divide the data into training and validation sets
df_train = data.sample(frac=0.7, random_state=42)  # 70% for training
df_train = df_train.reset_index(drop=True)
df_test = data.drop(df_train.index)  # remaining 30% for testing
df_test = df_test.reset_index(drop=True)
x_test = df_test.drop(columns=['time', 'event']).values

# ==============================================================================
# 2. Fit a Cox Proportional Hazards Model
# ==============================================================================
cph = CoxPHFitter()
cph.fit(df_train, duration_col='time', event_col='event')
year = 3  # Set the target year for calibration

# Predict the survival probability at the given time points
survs_cox = cph.predict_survival_function(x_test)
times = survs_cox.index.values
# draw the survival curves
fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(6, 2.5))
for i in range(20):
    ax[0].plot(times, survs_cox[i], label=f'Sample {i+1}', color=color_list[i % len(color_list)])
ax[0].set_ylabel('Survival Probability')
ax[0].set_ylim(-0.05, 1.05)
ax[0].set_xlabel('Time (days)')
ax[0].set_title('(a) Sampled Survival Curves')

eval = LifelinesEvaluator(survs_cox, df_test['time'].values, df_test['event'].values)

average_pred_surv = survs_cox.mean(axis=1)
km = KaplanMeierFitter().fit(df_test['time'], df_test['event'])
km_surv = km.survival_function_.values
km_times = km.survival_function_.index.values
ax[1].plot(times, average_pred_surv, label='Average prediction', color='skyblue', linewidth=2)
ax[1].plot(km_times, km_surv, label='Ideal (KM)', color='grey', linestyle='--', linewidth=2)
ax[1].set_ylabel('Survival Probability')
ax[1].set_xlabel('Time (days)')
ax[1].set_ylim(-0.05, 1.05)
ax[1].legend()
ax[1].set_title('(b) KM-calibration')

plt.tight_layout()
plt.savefig('km_calibration.png', dpi=400)


import datetime
import numpy as np
from dataclasses import field
from tqdm import trange

from SurvivalEVAL.NonparametricEstimator.SingleEvent import KaplanMeier, KaplanMeierArea


class KaplanMeierAreaOld(KaplanMeier):
    area_times: np.array = field(init=False)
    area_probabilities: np.array = field(init=False)
    area: np.array = field(init=False)

    def __post_init__(self, event_times, event_indicators):
        super().__post_init__(event_times, event_indicators)
        area_probabilities = np.append(1, self.survival_probabilities)
        area_times = np.append(0, self.survival_times)
        if self.survival_probabilities[-1] != 0:
            slope = (area_probabilities[-1] - 1) / area_times[-1]
            zero_survival = -1 / slope
            area_times = np.append(area_times, zero_survival)
            area_probabilities = np.append(area_probabilities, 0)

        area_diff = np.diff(area_times, 1)
        area_subs = area_diff * area_probabilities[0:-1]
        area_subs[-1] = area_subs[-1] / 2
        area = np.flip(np.flip(area_subs).cumsum())
        # area = np.flip(np.flip(area_diff * area_probabilities[0:-1]).cumsum())

        self.area_times = np.append(area_times, np.inf)
        self.area_probabilities = area_probabilities
        self.area = np.append(area, 0)

    def best_guess(self, censor_times: np.array):
        surv_prob = self.predict(censor_times)
        censor_indexes = np.digitize(censor_times, self.area_times)
        censor_indexes = np.where(
            censor_indexes == self.area_times.size + 1,
            censor_indexes - 1,
            censor_indexes,
        )
        censor_area = (
            self.area_times[censor_indexes] - censor_times
        ) * self.area_probabilities[censor_indexes - 1]
        censor_area += self.area[censor_indexes]
        return censor_times + censor_area / surv_prob


def km_mean(
        times: np.ndarray,
        survival_probabilities: np.ndarray
) -> float:
    """
    Calculate the mean of the Kaplan-Meier curve.

    Parameters
    ----------
    times: np.ndarray, shape = (n_samples, )
        Survival times for KM curve of the testing samples
    survival_probabilities: np.ndarray, shape = (n_samples, )
        Survival probabilities for KM curve of the testing samples

    Returns
    -------
    The mean of the Kaplan-Meier curve.
    """
    # calculate the area under the curve for each interval
    area_probabilities = np.append(1, survival_probabilities)
    area_times = np.append(0, times)
    km_linear_zero = -1 / ((area_probabilities[-1] - 1) / area_times[-1])
    if survival_probabilities[-1] != 0:
        area_times = np.append(area_times, km_linear_zero)
        area_probabilities = np.append(area_probabilities, 0)
    area_diff = np.diff(area_times, 1)
    # we are using trap rule
    # average_probabilities = (area_probabilities[0:-1] + area_probabilities[1:]) / 2
    # area = np.flip(np.flip(area_diff * average_probabilities).cumsum())
    # area = np.append(area, 0)
    area_subs = area_diff * area_probabilities[0:-1]
    area_subs[-1] = area_subs[-1] / 2
    area = np.flip(np.flip(area_subs).cumsum())

    # calculate the mean
    surv_prob = get_prob_at_zero(times, survival_probabilities)
    return area[0] / surv_prob


def get_prob_at_zero(
        times: np.ndarray,
        survival_probabilities: np.ndarray
) -> float:
    """
    Get the survival probability at time 0. Note that this function doesn't consider the interpolation.

    Parameters
    ----------
    times: np.ndarray, shape = (n_samples, )
        Survival times for KM curve of the testing samples
    survival_probabilities: np.ndarray, shape = (n_samples, )
        Survival probabilities for KM curve of the testing samples

    Returns
    -------
    The survival probability at time 0.
    """
    probability_index = np.digitize(0, times)
    probability = np.append(1, survival_probabilities)[probability_index]

    return probability


def PO_old(t_train, e_train, t_test, e_test, n_train, n_test, verbose):
    # ICML version
    start_time = datetime.datetime.now()
    km_model = KaplanMeierArea(t_train, e_train)
    pause_time_1 = datetime.datetime.now()
    print("Time elapsed for KM: ", pause_time_1 - start_time)
    best_guesses = np.empty(n_test)
    pasue_time_2 = datetime.datetime.now()
    sub_expect_time = km_model._compute_best_guess(0)
    pasue_time_3 = datetime.datetime.now()
    print("Time elapsed for sub_expect_time: ", pasue_time_3 - pasue_time_2)
    total_event_time = np.empty(n_train + 1)
    total_event_indicator = np.empty(n_train + 1)
    total_event_time[0:-1] = t_train
    total_event_indicator[0:-1] = e_train
    build_km_time = 0
    cal_mean_time = 0
    for i in trange(n_test, desc="Calculating surrogate times for MAE-PO", disable=not verbose):
        if e_test[i] == 1:
            best_guesses[i] = t_test[i]
        else:
            total_event_time[-1] = t_test[i]
            total_event_indicator[-1] = 0
            pasue_time_4 = datetime.datetime.now()
            total_km_model = KaplanMeierArea(total_event_time, total_event_indicator)
            pasue_time_5 = datetime.datetime.now()
            total_expect_time = total_km_model._compute_best_guess(0) # TODO: check why best_guess and _best_guess_revise give different results
            pasue_time_6 = datetime.datetime.now()
            best_guesses[i] = (n_train + 1) * total_expect_time - n_train * sub_expect_time
            build_km_time += (pasue_time_5 - pasue_time_4).total_seconds()
            cal_mean_time += (pasue_time_6 - pasue_time_5).total_seconds()

    print("Time elapsed for build_km_time: ", build_km_time)
    print("Time elapsed for cal_mean_time: ", cal_mean_time)
    return best_guesses

def PO(t_train, e_train, t_test, e_test, n_train, n_test, verbose):
    # the current survivalEVAL version, but slightly slower
    start_time = datetime.datetime.now()
    km_model = KaplanMeierArea(t_train, e_train)
    pause_time_1 = datetime.datetime.now()
    print("Time elapsed for KM: ", pause_time_1 - start_time)
    best_guesses = np.empty(n_test)
    pasue_time_2 = datetime.datetime.now()
    sub_expect_time = km_model.mean
    pasue_time_3 = datetime.datetime.now()
    print("Time elapsed for sub_expect_time: ", pasue_time_3 - pasue_time_2)
    total_event_time = np.empty(n_train + 1)
    total_event_indicator = np.empty(n_train + 1)
    total_event_time[0:-1] = t_train
    total_event_indicator[0:-1] = e_train
    build_km_time = 0
    cal_mean_time = 0
    for i in trange(n_test, desc="Calculating surrogate times for MAE-PO", disable=not verbose):
        if e_test[i] == 1:
            best_guesses[i] = t_test[i]
        else:
            total_event_time[-1] = t_test[i]
            total_event_indicator[-1] = 0
            pasue_time_4 = datetime.datetime.now()
            total_km_model = KaplanMeierArea(total_event_time, total_event_indicator)
            pasue_time_5 = datetime.datetime.now()
            total_expect_time = total_km_model.mean
            pasue_time_6 = datetime.datetime.now()
            best_guesses[i] = (n_train + 1) * total_expect_time - n_train * sub_expect_time
            build_km_time += (pasue_time_5 - pasue_time_4).total_seconds()
            cal_mean_time += (pasue_time_6 - pasue_time_5).total_seconds()

    print("Time elapsed for build_km_time: ", build_km_time)
    print("Time elapsed for cal_mean_time: ", cal_mean_time)
    return best_guesses


def PO_fast(t_train, e_train, t_test, e_test, n_train, n_test, verbose):
    # the current survivalEVAL version, with same speed, but slightly different results (using step func for area)
    start_time = datetime.datetime.now()
    km_model = KaplanMeierArea(t_train, e_train)
    pause_time_1 = datetime.datetime.now()
    print("Time elapsed for KM: ", pause_time_1 - start_time)
    best_guesses = np.empty(n_test)
    pasue_time_2 = datetime.datetime.now()
    # sub_expect_time = km_model.mean
    sub_expect_time = km_mean(km_model.survival_times, km_model.survival_probabilities)
    pasue_time_3 = datetime.datetime.now()
    print("Time elapsed for sub_expect_time: ", pasue_time_3 - pasue_time_2)
    # build_km_time = 0
    # cal_mean_time = 0
    for i in trange(n_test, desc="Calculating surrogate times for MAE-PO", disable=not verbose):
        if e_test[i] == 1:
            best_guesses[i] = t_test[i]
        else:
            pasue_time_4 = datetime.datetime.now()
            total_times, survival_probabilities = insert_km(km_model.survival_times.copy(), km_model.events.copy(), km_model.population_count.copy(),
                                                  t_test[i], 0)
            total_expect_time = km_mean(total_times, survival_probabilities)
            # total_expect_time = get_km_mean_total(km_model.survival_times.copy(), km_model.events.copy(), km_model.population_count.copy(),
            #                                       t_test[i], e_test[i])
            pasue_time_6 = datetime.datetime.now()
            best_guesses[i] = (n_train + 1) * total_expect_time - n_train * sub_expect_time
            # cal_mean_time += (pasue_time_6 - pasue_time_4).total_seconds()
    end_time = datetime.datetime.now()
    print("Time elapsed for PO_fast: ", end_time - start_time)
    return best_guesses


def PO_more_fast(t_train, e_train, t_test, e_test, n_train, n_test, verbose):
    # the fastest way, using dynamic programming
    start_time = datetime.datetime.now()
    km_model = KaplanMeierArea(t_train, e_train)
    sub_expect_time = km_mean(km_model.survival_times.copy(), km_model.survival_probabilities.copy())
    # sub_expect_time = km_model.mean

    events, population_counts = km_model.events.copy(), km_model.population_count.copy()
    times = km_model.survival_times.copy()

    unique_idx = np.where(events != 0)[0]
    if unique_idx[-1] != len(events) - 1:
        unique_idx = np.append(unique_idx, len(events) - 1)
    times = times[unique_idx]
    population_counts = population_counts[unique_idx]
    events = events[unique_idx]

    multiplier = 1 - events / population_counts
    multiplier_total = 1 - events / (population_counts + 1)
    best_guesses = t_test.copy().astype(float)

    # censor_times = t_test[e_test == 0]
    # censor_times = np.sort(censor_times)
    # insert_index = np.searchsorted(times, censor_times)
    #
    # for i in range(len(np.unique(insert_index)))

    for i in trange(n_test, desc="Calculating surrogate times for MAE-PO", disable=not verbose):
        if e_test[i] != 1:
            # TODO: no outliers
            total_multiplier = multiplier.copy()
            insert_index = np.searchsorted(times, t_test[i])
            total_multiplier[:insert_index+1] = multiplier_total[:insert_index+1]
            survival_probabilities = np.cumprod(total_multiplier)
            if insert_index == len(times):
                times_addition = np.append(times, t_test[i])
                survival_probabilities_addition = np.append(survival_probabilities, survival_probabilities[-1])
                total_expect_time = km_mean(times_addition, survival_probabilities_addition)
            else:
                total_expect_time = km_mean(times, survival_probabilities)
            best_guesses[i] = (n_train + 1) * total_expect_time - n_train * sub_expect_time
            # cal_mean_time += (pasue_time_6 - pasue_time_4).total_seconds()
    end_time = datetime.datetime.now()
    print("Time elapsed for PO_more_fast: ", end_time - start_time)
    return best_guesses


def insert_km(
        survival_times: np.ndarray,
        event_count: np.ndarray,
        as_risk_count: np.ndarray,
        new_t: float,
        new_e: int
) -> (np.ndarray, np.ndarray):
    # Find the index where new_t should be inserted
    insert_index = np.searchsorted(survival_times, new_t)

    # Check if new_t is already at the found index
    if insert_index < len(survival_times) and survival_times[insert_index] == new_t:
        # If new_t is already in the array, increment the event count
        event_count[insert_index] += new_e
        as_risk_count[:insert_index + 1] += 1
    else:
        # Insert new_t into
        survival_times = np.insert(survival_times, insert_index, new_t)
        event_count = np.insert(event_count, insert_index, new_e)
        if insert_index == survival_times.size - 1:
            as_risk_count = np.insert(as_risk_count, insert_index, 0)
        else:
            as_risk_count = np.insert(as_risk_count, insert_index, as_risk_count[insert_index])
        as_risk_count[:insert_index + 1] += 1

    event_ratios = 1 - event_count / as_risk_count
    survival_probabilities = np.cumprod(event_ratios)

    return survival_times, survival_probabilities


def margin_0(t_train, e_train, t_test, e_test, n_train, n_test, verbose):
    # the version that Haider created, and changing the last step to triangle
    start_time = datetime.datetime.now()
    km_model = KaplanMeierAreaOld(t_train, e_train)
    censor_times = t_test[e_test == 0]
    best_guesses = np.empty(n_test)
    best_guesses[e_test == 0] = km_model.best_guess(censor_times)
    best_guesses[e_test == 1] = t_test[e_test == 1]
    end_time = datetime.datetime.now()
    print("Time elapsed for margin_0: ", end_time - start_time)
    return best_guesses


def margin_1(t_train, e_train, t_test, e_test, n_train, n_test, verbose):
    # the version that used in the ICML paper
    start_time = datetime.datetime.now()
    km_model = KaplanMeierArea(t_train, e_train)
    censor_times = t_test[e_test == 0]
    best_guesses = np.empty(n_test)
    best_guesses[e_test == 0] = km_model.best_guess_revise(censor_times)
    best_guesses[e_test == 1] = t_test[e_test == 1]
    end_time = datetime.datetime.now()
    print("Time elapsed for margin_1: ", end_time - start_time)
    return best_guesses

def margin_2(t_train, e_train, t_test, e_test, n_train, n_test, verbose):
    # the new version, using trapezoid rule
    start_time = datetime.datetime.now()
    km_model = KaplanMeierArea(t_train, e_train)
    censor_times = t_test[e_test == 0]
    best_guesses = np.empty(n_test)
    best_guesses[e_test == 0] = km_model.best_guess(censor_times)
    best_guesses[e_test == 1] = t_test[e_test == 1]
    end_time = datetime.datetime.now()
    print("Time elapsed for margin_2: ", end_time - start_time)
    return best_guesses


if __name__ == "__main__":
    # get running times of each part of the code

    # first randomly generate some data
    verbose = True

    # test case #1
    # n_train = 100000
    # n_test = int(0.1 * n_train)
    # np.random.seed(10)
    #
    # # generate event times and event indicators
    # t_train = np.round(np.random.uniform(0, 5, n_train), 2)
    # e_train = np.random.binomial(1, 0.3, n_train)
    #
    # t_test = np.round(np.random.uniform(0, 10, n_test), 2)
    # e_test = np.random.binomial(1, 0.3, n_test)

    # test case #2
    # t_train = np.array([0, 1, 2, 3, 4, 5, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22,
    #                     26, 27, 28, 29, 30, 31, 32, 33, 34,  60, 61, 62, 63, 64, 65, 66, 67,
    #                     74, 75, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93,
    #                     98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111,
    #                     117, 118, 119, 120, 120, 120, 121, 121, 124, 125, 126, 127, 128, 129,
    #                     136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149,
    #                     155, 156, 157, 158, 159, 161, 182, 183, 186, 190, 191, 192, 192, 192,
    #                     193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 202, 203,
    #                     204, 202, 203, 204, 212, 213, 214, 215, 216, 217, 222, 223, 224])
    # e_train = np.array([1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0,
    #                     1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    #                     0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0,
    #                     0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
    #                     0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0,
    #                     1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0,
    #                     0, 0, 1, 1, 0, 0, 0, 0, 0, 0])
    # t_test = np.array([5, 10, 19, 31, 43, 59, 63, 75, 97, 113, 134, 151, 163, 176, 182, 195, 200, 210, 220, 250])
    # e_test = np.array([1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0])
    # n_train = t_train.size
    # n_test = t_test.size

    # test case #3
    t_train = np.load("../train_t.npy", allow_pickle=True)
    e_train = np.load("../train_e.npy", allow_pickle=True)
    t_test = np.load("../test_t.npy", allow_pickle=True)
    e_test = np.load("../test_e.npy", allow_pickle=True)
    n_train = t_train.size
    n_test = t_test.size

    # po0 = PO_old(t_train, e_train, t_test, e_test, n_train, n_test, verbose)
    # po1 = PO(t_train, e_train, t_test, e_test, n_train, n_test, verbose)
    po2 = PO_fast(t_train, e_train, t_test, e_test, n_train, n_test, verbose)
    # po3 = PO_more_fast(t_train, e_train, t_test, e_test, n_train, n_test, verbose)
    # print("Does the three PO methods give the same results? ", np.all(po1 == po2) and np.all(po1 == po3))
    # m0 = margin_0(t_train, e_train, t_test, e_test, n_train, n_test, verbose)
    # m1 = margin_1(t_train, e_train, t_test, e_test, n_train, n_test, verbose)
    # m2 = margin_2(t_train, e_train, t_test, e_test, n_train, n_test, verbose)
    # print("Does the three margin methods give the same results? ", np.all(m0 == m1) and np.all(m0 == m2))



import numpy as np
from sklearn.metrics import roc_auc_score


def auc(
        predict_probs: np.ndarray,
        event_times: np.ndarray,
        event_indicators: np.ndarray,
        target_time: float = None
) -> float:
    """
    Calculate the Area Under the Curve (AUC) for the survival model.

    Parameters
    ----------
    predict_probs: np.ndarray
        The predicted survival probabilities
    event_times: np.ndarray
        The event or censoring times for the test data
    event_indicators: np.ndarray
        The binary indicators of whether the event occurred (1) or was censored (0)
    target_time: float, optional
        The specific time point at which to calculate the AUC. If not specified, the median of the event times is used.

    Returns
    -------
    auc: float
        The AUC value calculated at the specified target time.
    """
    # if the target time is not specified, then we use the median of the event times
    if target_time is None:
        target_time = np.median(event_times)

    # for censored data, if the censor time is earlier than the target time,
    # (since we cannot observe the real status at the target time)
    # then we just exclude its prediction and observation from the calculation
    exclude_indicators = np.logical_and(event_times < target_time, event_indicators == 0)
    event_times = event_times[~exclude_indicators]
    predict_probs = predict_probs[~exclude_indicators]

    # get the binary status of the test data, given the target time
    binary_status = (event_times <= target_time).astype(int)

    # check if the binary status is all zeros or all ones
    if np.all(binary_status == 0) or np.all(binary_status == 1):
        raise ValueError(f"Survival status is all zeros or all ones at time: {target_time}, AUC cannot be computed.")

    # computing the AUC, given the predicted probabilities and the binary status
    risks = 1 - predict_probs
    return roc_auc_score(binary_status, risks)


if __name__ == '__main__':
    # test the AUC function
    probs = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.2, 0.8, 0.9])
    times = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
    indicators = np.array([1, 1, 0, 1, 1, 1, 1, 1, 1])
    target_t = 5

    print(auc(probs, times, indicators, target_t))


#%% Discrimination vs Calibration
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from scipy.integrate import cumtrapz
import math
import seaborn as sns
import matplotlib.patches as patches
import matplotlib.lines as mlines


color_list = sns.color_palette("colorblind", 10)
custom_params = {"axes.spines.right": False, "axes.spines.top": False,}
sns.set_theme(style="white", rc=custom_params)

# sns.set_theme(style="whitegrid")
# plt.style.use('fivethirtyeight')
# plt.rcParams["figure.figsize"] = (5, 5)
plt.rc('axes', titlesize=10)     # fontsize of the axes title
plt.rc('axes', labelsize=10)    # fontsize of the x and y labels
# plt.rc('xtick', labelsize=10)    # fontsize of the tick labels
# plt.rc('ytick', labelsize=10)    # fontsize of the tick labels
plt.rc('legend', fontsize=10)    # legend fontsize
plt.rc('figure', titlesize=10)  # fontsize of the figure title
height_unit = 2.4
width_unit = 2.8
plt.rcParams['font.size'] = 12


def weibull_pdf(x, scale, shape):
    return (shape / scale) * (x / scale)**(shape - 1) * np.exp(-(x / scale)**shape)


def gamma_pdf(x, scale, shape):
    return (x ** (shape - 1) * np.exp(-x / scale)) / (math.gamma(shape) * scale ** shape)


def lognormal_pdf(x, mu, sigma):
    return (1 / (x * sigma * np.sqrt(2 * np.pi))) * np.exp(-(np.log(x) - mu) ** 2 / (2 * sigma ** 2))


e_1 = 2.7
e_2 = 1.5
e_3 = 0.6
e_4 = 2.1
end_time = 3
N = 100000

# ISD figures and event times
fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(width_unit*3, height_unit))

# good Cindex but poor calibration
s1 = 0.5 * np.random.weibull(a=1, size=int(0.55*N))   # generate your data sample with N elements
s2 = 2.5 * np.random.weibull(a=7, size=int(0.45*N))
sample_all = np.append(s1, s2)
kde = gaussian_kde(sample_all)
# these are the values over which your kernel will be evaluated
dist_space = np.arange(0, end_time, 0.001)
p_1 = kde(dist_space)
# plt.plot( dist_space, pdf )
cdf_1 = cumtrapz(p_1, dist_space, initial=0)
death_idx_1 = int(e_1/0.001)
# ab_1 = AnnotationBbox(death_imagebox, (e_1, 1 - cdf_1[death_idx_1]), frameon = False)

p_2 = weibull_pdf(dist_space, 0.5, 1.1)
cdf_2 = cumtrapz(p_2, dist_space, initial=0)
death_idx_2 = int(e_2/0.001)
p_3 = gamma_pdf(dist_space, 0.2, 1)
cdf_3 = cumtrapz(p_3, dist_space, initial=0)
death_idx_3 = int(e_3/0.001)
p_4 = lognormal_pdf(dist_space, 0.002, 0.1)
p_4[0] = 0
cdf_4 = cumtrapz(p_4, dist_space, initial=0)
death_idx_4 = int(e_4/0.001)


axes[0].plot(dist_space, 1 - cdf_1, c=color_list[0], linewidth=2, clip_on=False)
axes[0].plot(dist_space, 1 - cdf_2, c=color_list[1], linewidth=2, clip_on=False)
axes[0].plot(dist_space, 1 - cdf_3, c=color_list[2], linewidth=2, clip_on=False)
axes[0].plot(dist_space, 1 - cdf_4, c=color_list[3], linewidth=2, clip_on=False)
axes[0].plot(e_1, 1 - cdf_1[death_idx_1], marker='*', label='A', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[0].plot(e_2, 1 - cdf_2[death_idx_2], marker='*', label='B', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[0].plot(e_3, 1 - cdf_3[death_idx_3], marker='*', label='C', markersize=12, c=color_list[2], clip_on=False, zorder=10)
axes[0].plot(e_4, 1 - cdf_4[death_idx_4], marker='*', label='D', markersize=12, c=color_list[3], clip_on=False, zorder=10)
# axes.axhline(0.5, ls='dashdot', c='grey', label='50% Prob.')

axes[0].grid(axis='y')
axes[0].set_ylim([-0.0, 1.0])
axes[0].set_xlim([0, end_time])
axes[0].set_xlabel('Days')
axes[0].set_ylabel('Survival Probability')

# poor C-index but good calibration
s1 = 0.4 * np.random.weibull(a=1, size=int(0.7*N))   # generate your data sample with N elements
s2 = 3.3 * np.random.weibull(a=7, size=int(0.3*N))
sample_all = np.append(s1, s2)
kde = gaussian_kde(sample_all)
# these are the values over which your kernel will be evaluated
dist_space = np.arange(0, end_time, 0.001)
p_1 = kde(dist_space)
# plt.plot( dist_space, pdf )
cdf_1 = cumtrapz(p_1, dist_space, initial=0)
death_idx_1 = int(e_1/0.001)
# ab_1 = AnnotationBbox(death_imagebox, (e_1, 1 - cdf_1[death_idx_1]), frameon = False)

p_2 = weibull_pdf(dist_space, 2, 2)
cdf_2 = cumtrapz(p_2, dist_space, initial=0)
death_idx_2 = int(e_2/0.001)
p_3 = gamma_pdf(dist_space, 1.2, 2.7)
cdf_3 = cumtrapz(p_3, dist_space, initial=0)
death_idx_3 = int(e_3/0.001)
p_4 = lognormal_pdf(dist_space, 0.1, 0.5)
p_4[0] = 0
cdf_4 = cumtrapz(p_4, dist_space, initial=0)
death_idx_4 = int(e_4/0.001)

axes[1].plot(dist_space, 1 - cdf_1, c=color_list[0], linewidth=2, clip_on=False)
axes[1].plot(dist_space, 1 - cdf_2, c=color_list[1], linewidth=2, clip_on=False)
axes[1].plot(dist_space, 1 - cdf_3, c=color_list[2], linewidth=2, clip_on=False)
axes[1].plot(dist_space, 1 - cdf_4, c=color_list[3], linewidth=2, clip_on=False)
axes[1].plot(e_1, 1 - cdf_1[death_idx_1], marker='*', label='A', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[1].plot(e_2, 1 - cdf_2[death_idx_2], marker='*', label='B', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[1].plot(e_3, 1 - cdf_3[death_idx_3], marker='*', label='C', markersize=12, c=color_list[2], clip_on=False, zorder=10)
axes[1].plot(e_4, 1 - cdf_4[death_idx_4], marker='*', label='D', markersize=12, c=color_list[3], clip_on=False, zorder=10)
# axes.axhline(0.5, ls='dashdot', c='grey', label='50% Prob.')

axes[1].grid(axis='y')
axes[1].set_ylim([-0.0, 1.0])
axes[1].set_xlim([0, end_time])
axes[1].set_yticks([0, 0.25, 0.5, 0.75, 1])
axes[1].set_xlabel('Days')
axes[1].set_ylabel('Survival Probability')


# good C-index and calibration
s1 = 0.5 * np.random.weibull(a=1, size=int(0.23*N))   # generate your data sample with N elements
s2 = 3.2 * np.random.weibull(a=7, size=int(0.77*N))
sample_all = np.append(s1, s2)
kde = gaussian_kde(sample_all)
# these are the values over which your kernel will be evaluated
dist_space = np.arange(0, end_time, 0.001)
p_1 = kde(dist_space)
# plt.plot( dist_space, pdf )
cdf_1 = cumtrapz(p_1, dist_space, initial=0)
death_idx_1 = int(e_1/0.001)
# ab_1 = AnnotationBbox(death_imagebox, (e_1, 1 - cdf_1[death_idx_1]), frameon = False)

p_2 = weibull_pdf(dist_space, 1.5, 1.5)
cdf_2 = cumtrapz(p_2, dist_space, initial=0)
death_idx_2 = int(e_2/0.001)
p_3 = gamma_pdf(dist_space, 0.15, 2)
cdf_3 = cumtrapz(p_3, dist_space, initial=0)
death_idx_3 = int(e_3/0.001)
p_4 = lognormal_pdf(dist_space, 0.9, 0.15)
p_4[0] = 0
cdf_4 = cumtrapz(p_4, dist_space, initial=0)
death_idx_4 = int(e_4/0.001)

axes[2].plot(dist_space, 1 - cdf_1, c=color_list[0], linewidth=2, clip_on=False)
axes[2].plot(dist_space, 1 - cdf_2, c=color_list[1], linewidth=2, clip_on=False)
axes[2].plot(dist_space, 1 - cdf_3, c=color_list[2], linewidth=2, clip_on=False)
axes[2].plot(dist_space, 1 - cdf_4, c=color_list[3], linewidth=2, clip_on=False)
axes[2].plot(e_1, 1 - cdf_1[death_idx_1], marker='*', label='A', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[2].plot(e_2, 1 - cdf_2[death_idx_2], marker='*', label='B', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[2].plot(e_3, 1 - cdf_3[death_idx_3], marker='*', label='C', markersize=12, c=color_list[2], clip_on=False, zorder=10)
axes[2].plot(e_4, 1 - cdf_4[death_idx_4], marker='*', label='D', markersize=12, c=color_list[3], clip_on=False, zorder=10)
# axes.axhline(0.5, ls='dashdot', c='grey', label='50% Prob.')

axes[2].grid(axis='y')
axes[2].set_ylim([-0.0, 1.0])
axes[2].set_xlim([0, end_time])
axes[2].set_xlabel('Days')
axes[2].set_ylabel('Survival Probability')

# axes[0].legend(loc='lower center', bbox_to_anchor=(0.5, 1.02), fancybox=True,
#              ncols=2,)
plt.tight_layout()
plt.savefig(f'figs/illustration/disc_vs_cal.png', dpi=400, transparent=True)



import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from scipy.integrate import cumulative_trapezoid
import math
import seaborn as sns
import matplotlib.patches as patches
import matplotlib.lines as mlines

# import matplotlib.image as image
# from matplotlib.offsetbox import (OffsetImage, AnnotationBbox)
#
# death_img_file = "Figs/illustration/death.png"
# death_img = image.imread(death_img_file)
# death_imagebox = OffsetImage(death_img, zoom = 0.02)

color_list = sns.color_palette("deep", 6)


plt.style.use('seaborn-v0_8-paper')
plt.rcParams["figure.figsize"] = (9.6, 8)
plt.rc('axes', titlesize=9.5)     # fontsize of the axes title
plt.rc('axes', labelsize=9.5)    # fontsize of the x and y labels
# plt.rc('xtick', labelsize=10)    # fontsize of the tick labels
# plt.rc('ytick', labelsize=10)    # fontsize of the tick labels
plt.rc('legend', fontsize=9.5)    # legend fontsize
plt.rc('figure', titlesize=9.5)  # fontsize of the figure title

plt.rcParams['font.size'] = 12


def weibull_pdf(x, scale, shape):
    return (shape / scale) * (x / scale)**(shape - 1) * np.exp(-(x / scale)**shape)

def gamma_pdf(x, scale, shape):
    return (x ** (shape - 1) * np.exp(-x / scale)) / (math.gamma(shape) * scale ** shape)

def lognormal_pdf(x, mu, sigma):
    return (1 / (x * sigma * np.sqrt(2 * np.pi))) * np.exp(-(np.log(x) - mu) ** 2 / (2 * sigma ** 2))

def heaviside(x, time):
    function = np.zeros_like(x)
    function[x < time] = 1
    return function

e_1 = 1.5
e_2 = 1.8
e_3 = 0.8
e_4 = 2.15
end_time = 3
N = 100000
s1 = 0.5 * np.random.weibull(a=1, size=int(0.4*N))   # generate your data sample with N elements
s2 = 2.8 * np.random.weibull(a=7, size=int(0.6*N))
sample_all = np.append(s1, s2)
kde = gaussian_kde(sample_all)
# these are the values over which your kernel will be evaluated
dist_space = np.arange(0, end_time, 0.001)
p_1 = kde(dist_space)
# plt.plot( dist_space, pdf )
cdf_1 = cumulative_trapezoid(p_1, dist_space, initial=0)
death_idx_1 = int(e_1/0.001)
# ab_1 = AnnotationBbox(death_imagebox, (e_1, 1 - cdf_1[death_idx_1]), frameon = False)

p_2 = weibull_pdf(dist_space, 1.1, 2)
cdf_2 = cumulative_trapezoid(p_2, dist_space, initial=0)
death_idx_2 = int(e_2/0.001)
p_3 = gamma_pdf(dist_space, 0.6, 3)
cdf_3 = cumulative_trapezoid(p_3, dist_space, initial=0)
death_idx_3 = int(e_3/0.001)
p_4 = lognormal_pdf(dist_space, 0.7, 0.2)
p_4[0] = 0
cdf_4 = cumulative_trapezoid(p_4, dist_space, initial=0)
death_idx_4 = int(e_4/0.001)


fig, axes = plt.subplots(nrows=3, ncols=3)

# subfigure 0, concordance index
axes[0, 0].axhline(0.5, ls='dashdot', c='grey')
# axes[0, 0].axvline(e_1, ls='--', c=color_list[0], alpha=0.5)
# axes[0, 0].axvline(e_2, ls='--', c=color_list[1], alpha=0.5)
axes[0, 0].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
axes[0, 0].plot(dist_space, 1 - cdf_2, label='Patient B', c=color_list[1])
axes[0, 0].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[0, 0].plot(e_2, -0.05, marker='*', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[0, 0].plot(dist_space[np.where(cdf_1>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[0])
axes[0, 0].plot(dist_space[np.where(cdf_2>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[1])
# axes[0, 0].text(-0.05, 0.15, 'True  Risk', fontsize=10, horizontalalignment='left', verticalalignment="center")
axes[0, 0].text(-0.05, 0.05, 'Pred. Order', fontsize=9, horizontalalignment='left', verticalalignment="center")
# axes[0, 0].plot(1, 0.15, marker='*', markersize=9, c=color_list[0])
# axes[0, 0].plot(1.2, 0.15, marker='*', markersize=9, c=color_list[1])
axes[0, 0].plot(1, 0.05, marker='v', markersize=7.5, c=color_list[1])
axes[0, 0].plot(1.2, 0.05, marker='v', markersize=7.5, c=color_list[0])
axes[0, 0].grid(False)
axes[0, 0].set_ylim([-0.05, 1.05])
axes[0, 0].set_yticks([0, 0.5, 1])
axes[0, 0].set_title('(a) C-index', weight='bold')
axes[0, 0].set_xlabel('Year')
axes[0, 0].set_ylabel('Survival Probability')


# subfigure 1, l1 loss
axes[0, 1].axhline(0.5, ls='dashdot', c='grey')
axes[0, 1].axvline(e_1, ls='--', c=color_list[0], alpha=0.5)
axes[0, 1].axvline(dist_space[np.where(cdf_1 > 0.5)[0][0]], ls='--', c=color_list[0], alpha=0.5)
axes[0, 1].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
axes[0, 1].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], label=r'True $t_i$', clip_on=False, zorder=10)
axes[0, 1].plot(dist_space[np.where(cdf_1>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[0], label=r'Predict $\hat{t}_i$')
curve_arrow = patches.FancyArrowPatch((e_1 - 0.04, 0.35), (dist_space[np.where(cdf_1 > 0.5)[0][0]] + 0.04, 0.35), connectionstyle="arc3,rad=0", arrowstyle='<->', mutation_scale=10, color='black', linewidth=1.5)
axes[0, 1].add_patch(curve_arrow)
axes[0, 1].grid(False)
axes[0, 1].set_ylim([-0.05, 1.05])
axes[0, 1].set_yticks([0, 0.5, 1])
axes[0, 1].set_title('(b) MAE / MSE', weight='bold')
axes[0, 1].set_xlabel('Year')
axes[0, 1].set_ylabel('Survival Probability')

# subfigure 3, log-rank test
# axes[0, 2].axis('off')
axes[0, 2].axhline(0.5, ls='dashdot', c='grey')
axes[0, 2].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
axes[0, 2].plot(dist_space, 1 - cdf_2, label='Patient B', c=color_list[1])
axes[0, 2].plot(dist_space, 1 - cdf_3, label='Patient C', c=color_list[2])
axes[0, 2].plot(dist_space, 1 - cdf_4, label='Patient D', c=color_list[3])
axes[0, 2].plot(dist_space[np.where(cdf_1>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[0])
axes[0, 2].plot(dist_space[np.where(cdf_2>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[1])
axes[0, 2].plot(dist_space[np.where(cdf_3>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[2])
axes[0, 2].plot(dist_space[np.where(cdf_4>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[3])
axes[0, 2].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[0, 2].plot(e_2, -0.05, marker='*', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[0, 2].plot(e_3, -0.05, marker='*', markersize=12, c=color_list[2], clip_on=False, zorder=10)
axes[0, 2].plot(e_4, -0.05, marker='*', markersize=12, c=color_list[3], clip_on=False, zorder=10)
# axes[0, 2].text(2.39, 0.8, '2 events in \n [0.5, 1]', fontsize=9, horizontalalignment='center')
# axes[0, 2].text(0.6, 0.15, '2 events in \n [0, 0.5)', fontsize=9, horizontalalignment='center')
group_1 = patches.FancyBboxPatch((.8, 0.5), 1.6, 0.001, linewidth=0.5, edgecolor='none', facecolor=color_list[4], alpha=0.5, boxstyle=patches.BoxStyle("round", pad=0.07))
axes[0, 2].add_patch(group_1)
group_2 = patches.FancyBboxPatch((0.7, -0.05), 1.6, 0.001, linewidth=0.5, edgecolor='none', zorder=8, facecolor=color_list[5], alpha=0.5, clip_on=False, boxstyle=patches.BoxStyle("round", pad=0.07))
axes[0, 2].add_patch(group_2)

# add curve arrow between the two groups
curve_arrow = patches.FancyArrowPatch((0.7, 0.5), (0.6, -0.05), connectionstyle="arc3,rad=0.3", arrowstyle='<->', mutation_scale=10, color='black', linewidth=1.5)
axes[0, 2].add_patch(curve_arrow)
axes[0, 2].grid(False)
# axes[0, 2].spines.set_color('black')
axes[0, 2].set_ylim([-0.05, 1.05])
axes[0, 2].set_yticks([0, 0.5, 1])
axes[0, 2].set_title('(c) Log-rank test', weight='bold')
axes[0, 2].set_xlabel('Year')
axes[0, 2].set_ylabel('Survival Probability')


# subfigure 4, AUC
axes[1, 0].axvline(1.65, ls='--', c=color_list[5])
axes[1, 0].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
axes[1, 0].plot(dist_space, 1 - cdf_2, label='Patient B', c=color_list[1])
axes[1, 0].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[1, 0].plot(e_2, -0.05, marker='*', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[1, 0].plot(1.65, 1 - cdf_1[1650], marker='v', markersize=10, c=color_list[0])
axes[1, 0].plot(1.65, 1 - cdf_2[1650], marker='v', markersize=10, c=color_list[1])
# axes[1, 0].text(-0.05, 0.15, 'True  Risk', fontsize=10, horizontalalignment='left', verticalalignment="center")
axes[1, 0].text(-0.05, 0.05, 'Pred. Order', fontsize=9, horizontalalignment='left', verticalalignment="center")
# axes[1, 0].plot(1, 0.15, marker='*', markersize=9, c=color_list[0])
# axes[1, 0].plot(1.2, 0.15, marker='*', markersize=9, c=color_list[1])
axes[1, 0].plot(1, 0.05, marker='v', markersize=7.5, c=color_list[1])
axes[1, 0].plot(1.2, 0.05, marker='v', markersize=7.5, c=color_list[0])
axes[1, 0].grid(False)
axes[1, 0].set_ylim([-0.05, 1.05])
axes[1, 0].set_yticks([0, 0.5, 1])
axes[1, 0].set_title(r'(d) AUC ($t^*=1.65$)', weight='bold')
axes[1, 0].set_xlabel('Year')
axes[1, 0].set_ylabel('Survival Probability')


# subfigure 5, Brier Score
axes[1, 1].axvline(1.65, ls='--', c=color_list[5])
axes[1, 1].axhline(0, ls='--', c=color_list[0], alpha=0.5)
axes[1, 1].axhline(1 - cdf_1[int(1.65*1000)], ls='--', c=color_list[0], alpha=0.5)
axes[1, 1].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
# axes[1, 1].plot(dist_space[:death_idx_1], 1 - cdf_1[:death_idx_1], ls='-', label='Patient A', c=color_list[0])
# axes[1, 1].plot(dist_space[death_idx_1:], 1 - cdf_1[death_idx_1:], ls=(0, (1, 1)), label='Patient A', c=color_list[0])
axes[1, 1].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[1, 1].plot(1.65, 1 - cdf_1[1650], marker='v', markersize=10, c=color_list[0])
curve_arrow = patches.FancyArrowPatch((1.35, - 0.03), (1.35, 1 - cdf_1[int(1.65*1000)] + 0.03), connectionstyle="arc3,rad=0", arrowstyle='<->', mutation_scale=10, color='black', linewidth=1.5)
axes[1, 1].add_patch(curve_arrow)

axes[1, 1].grid(False)
axes[1, 1].set_ylim([-0.05, 1.05])
axes[1, 1].set_yticks([0, 0.5, 1])
axes[1, 1].set_title(r'(e) Brier score ($t^*=1.65$)', weight='bold')
axes[1, 1].set_xlabel('Year')
axes[1, 1].set_ylabel('Survival Probability')


# subfigure 6, 1-calibration
axes[1, 2].axvline(1.65, ls='--', c=color_list[5])
axes[1, 2].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
axes[1, 2].plot(dist_space, 1 - cdf_2, label='Patient B', c=color_list[1])
axes[1, 2].plot(dist_space, 1 - cdf_3, label='Patient C', c=color_list[2])
axes[1, 2].plot(dist_space, 1 - cdf_4, label='Patient D', c=color_list[3])
axes[1, 2].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[1, 2].plot(e_2, -0.05, marker='*', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[1, 2].plot(e_3, -0.05, marker='*', markersize=12, c=color_list[2], clip_on=False, zorder=10)
axes[1, 2].plot(e_4, -0.05, marker='*', markersize=12, c=color_list[3], clip_on=False, zorder=10)
axes[1, 2].plot(1.65, 1 - cdf_1[1650], marker='v', markersize=10, c=color_list[0])
axes[1, 2].plot(1.65, 1 - cdf_2[1650], marker='v', markersize=10, c=color_list[1])
axes[1, 2].plot(1.65, 1 - cdf_3[1650], marker='v', markersize=10, c=color_list[2])
axes[1, 2].plot(1.65, 1 - cdf_4[1650], marker='v', markersize=10, c=color_list[3])
group_1 = patches.FancyBboxPatch((1.61, 0.12), 0.1, 0.75, linewidth=1, edgecolor='none', facecolor=color_list[4], alpha=0.5, boxstyle=patches.BoxStyle("round", pad=0.07))
axes[1, 2].add_patch(group_1)
group_2 = patches.FancyBboxPatch((0.7, -0.05), 1.6, 0.001, linewidth=0.5, edgecolor='none', zorder=8, facecolor=color_list[5], alpha=0.5, clip_on=False, boxstyle=patches.BoxStyle("round", pad=0.07))
axes[1, 2].add_patch(group_2)

# add curve arrow between the two groups, arrow arc face up
curve_arrow = patches.FancyArrowPatch((2.2, 0.02), (1.75, 0.3),  connectionstyle="arc3,rad=0.3", arrowstyle='<->', mutation_scale=10, color='black', linewidth=1.5)
axes[1, 2].add_patch(curve_arrow)

# group_2 = patches.FancyBboxPatch((1.645, 0.65), 0.01, 0.20, linewidth=1, edgecolor='none', facecolor=color_list[4], alpha=0.5, boxstyle=patches.BoxStyle("round", pad=0.07))
# axes[1, 2].add_patch(group_2)
# axes[1, 2].text(2.1, 0.82, 'Two event', fontsize=9, horizontalalignment='center')
# axes[1, 2].text(2.1, 0.09, 'Group 2', fontsize=9, horizontalalignment='center')
axes[1, 2].text(2.42, 0.87, 'Avg. pred. prob. \n= 51.4%', fontsize=9, horizontalalignment='center')
axes[1, 2].text(0.6, 0.05, '2/4 events', fontsize=9, horizontalalignment='center')

axes[1, 2].grid(False)
axes[1, 2].set_ylim([-0.05, 1.05])
axes[1, 2].set_yticks([0, 0.5, 1])
axes[1, 2].set_title(r'(f) 1-calibration ($t^*=1.65$)', weight='bold')
axes[1, 2].set_xlabel('Year')
axes[1, 2].set_ylabel('Survival Probability')


# subfigure 7, td-C-index
# axes[2, 0].axhline(0.5, ls='dashdot', c='grey')
axes[2, 0].axvline(e_1, ls='--', c=color_list[0], alpha=0.5)
# axes[2, 0].axvline(e_2, ls='--', c=color_list[1], alpha=0.5)
axes[2, 0].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
axes[2, 0].plot(dist_space, 1 - cdf_2, label='Patient B', c=color_list[1])
axes[2, 0].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[2, 0].plot(e_2, -0.05, marker='*', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[2, 0].plot(e_1, 1 - cdf_1[int(e_1*1000)], marker='v', markersize=10, c=color_list[0])
axes[2, 0].plot(e_1, 1 - cdf_2[int(e_1*1000)], marker='v', markersize=10, c=color_list[1])
# axes[2, 0].plot(dist_space[np.where(cdf_1>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[0])
# axes[2, 0].plot(dist_space[np.where(cdf_2>0.5)[0][0]], 0.5, marker='v', markersize=10, c=color_list[1])
# axes[2, 0].text(-0.05, 0.15, 'True  Risk', fontsize=10, horizontalalignment='left', verticalalignment="center")
axes[2, 0].text(-0.05, 0.05, 'Pred. Order', fontsize=9, horizontalalignment='left', verticalalignment="center")
# axes[2, 0].plot(1, 0.15, marker='*', markersize=9, c=color_list[0])
# axes[2, 0].plot(1.2, 0.15, marker='*', markersize=9, c=color_list[1])
axes[2, 0].plot(1, 0.05, marker='v', markersize=7.5, c=color_list[1])
axes[2, 0].plot(1.2, 0.05, marker='v', markersize=7.5, c=color_list[0])
axes[2, 0].grid(False)
axes[2, 0].set_ylim([-0.05, 1.05])
axes[2, 0].set_yticks([0, 0.5, 1])
axes[2, 0].set_title('(g) Time-dependent C-index', weight='bold')
axes[2, 0].set_xlabel('Year')
axes[2, 0].set_ylabel('Survival Probability')


# subfigure 8, IBS
axes[2, 1].axvline(e_1, ls='--', c=color_list[0], alpha=0.5)
axes[2, 1].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
axes[2, 1].fill_between(dist_space, heaviside(dist_space, e_1), 1 - cdf_1, alpha=0.2, color='grey')
axes[2, 1].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[2, 1].text(1., 0.83, r'$\alpha$', fontsize=9, horizontalalignment='center')
axes[2, 1].text(2., 0.3, r'$\beta$', fontsize=9, horizontalalignment='center')
axes[2, 1].grid(False)
axes[2, 1].set_ylim([-0.05, 1.05])
axes[2, 1].set_yticks([0, 0.5, 1])
axes[2, 1].set_title('(h) Integrated Brier score', weight='bold')
axes[2, 1].set_xlabel('Year')
axes[2, 1].set_ylabel('Survival Probability')


# subfigure 9, d-calibration
axes[2, 2].axhline(0.5, ls='dashdot', c='grey')
axes[2, 2].plot(dist_space, 1 - cdf_1, label='Patient A', c=color_list[0])
axes[2, 2].plot(dist_space, 1 - cdf_2, label='Patient B', c=color_list[1])
axes[2, 2].plot(dist_space, 1 - cdf_3, label='Patient C', c=color_list[2])
axes[2, 2].plot(dist_space, 1 - cdf_4, label='Patient D', c=color_list[3])
axes[2, 2].plot(e_1, 1 - cdf_1[death_idx_1], marker='v', markersize=10, c=color_list[0])
axes[2, 2].plot(e_2, 1 - cdf_2[death_idx_2], marker='v', markersize=10, c=color_list[1])
axes[2, 2].plot(e_3, 1 - cdf_3[death_idx_3], marker='v', markersize=10, c=color_list[2])
axes[2, 2].plot(e_4, 1 - cdf_4[death_idx_4], marker='v', markersize=10, c=color_list[3])
axes[2, 2].plot(e_1, -0.05, marker='*', markersize=12, c=color_list[0], clip_on=False, zorder=10)
axes[2, 2].plot(e_2, -0.05, marker='*', markersize=12, c=color_list[1], clip_on=False, zorder=10)
axes[2, 2].plot(e_3, -0.05, marker='*', markersize=12, c=color_list[2], clip_on=False, zorder=10)
axes[2, 2].plot(e_4, -0.05, marker='*', markersize=12, c=color_list[3], clip_on=False, zorder=10)
axes[2, 2].text(2.39, 0.8, '2 events in \n [0.5, 1]', fontsize=9, horizontalalignment='center')
axes[2, 2].text(0.6, 0.15, '2 events in \n [0, 0.5)', fontsize=9, horizontalalignment='center')

axes[2, 2].grid(False)
# axes[2, 2].spines.set_color('black')
axes[2, 2].set_ylim([-0.05, 1.05])
axes[2, 2].set_yticks([0, 0.5, 1])
axes[2, 2].set_title('(i) D-calibration', weight='bold')
axes[2, 2].set_xlabel('Year')
axes[2, 2].set_ylabel('Survival Probability')

lines, labels = axes[2, 2].get_legend_handles_labels()
black_line = mlines.Line2D([], [], color='black', linestyle='-', label=r'Predicted $\widehat{S}(t\mid \bf{x}_i)$')
black_star = mlines.Line2D([], [], color='black', marker='*', linestyle='None',
                          markersize=12, label=r'True $t_i$')
# black_trangle = mlines.Line2D([], [], color='black', marker='v', linestyle='None',
#                           markersize=10, label=r'Prediction')
color_A = mlines.Line2D([], [], color=color_list[0], marker='s', linestyle='None',
                            markersize=12, label='A', markerfacecolor=color_list[0])
color_B = mlines.Line2D([], [], color=color_list[1], marker='s', linestyle='None',
                            markersize=12, label='B', markerfacecolor=color_list[1])
color_C = mlines.Line2D([], [], color=color_list[2], marker='s', linestyle='None',
                            markersize=12, label='C', markerfacecolor=color_list[2])
color_D = mlines.Line2D([], [], color=color_list[3], marker='s', linestyle='None',
                            markersize=12, label='D', markerfacecolor=color_list[3])
lines = []
lines.append(black_star)
lines.append(black_line)
# lines.append(black_trangle)
lines.append(color_A)
lines.append(color_B)
lines.append(color_C)
lines.append(color_D)

labels = []
labels.append(r'True time $t_i$')
labels.append(r'Prediction $\hat{S}(t\mid x_i)$')
# labels.append('Prediction')
labels.append('A')
labels.append('B')
labels.append('C')
labels.append('D')
lgd = fig.legend(lines, labels, bbox_to_anchor=(0.88, 0.02), prop={'size': 9.5}, ncol=7)

# plt.axhline(0.5, ls='--', c='grey')
# plt.legend()
plt.tight_layout()

# plt.show()
# plt.savefig('all_metrics.pdf', dpi=400, bbox_extra_artists=(lgd,), bbox_inches='tight')
plt.savefig('all_metrics.png', dpi=400, bbox_extra_artists=(lgd,), bbox_inches='tight',
            transparent=False)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import seaborn as sns
import lifelines
from lifelines import CoxPHFitter

from SurvivalEVAL import LifelinesEvaluator
from SurvivalEVAL.NonparametricEstimator.SingleEvent import  NelsonAalen

color_list = sns.color_palette(n_colors=6)

plt.style.use('seaborn-v0_8-paper')
plt.rcParams["figure.figsize"] = (8, 3.5)
plt.rc('axes', titlesize=9.5)     # fontsize of the axes title
plt.rc('axes', labelsize=9.5)    # fontsize of the x and y labels
# plt.rc('xtick', labelsize=10)    # fontsize of the tick labels
# plt.rc('ytick', labelsize=10)    # fontsize of the tick labels
plt.rc('legend', fontsize=9.5)    # legend fontsize
plt.rc('figure', titlesize=9.5)  # fontsize of the figure title

plt.rcParams['font.size'] = 12


# ==============================================================================
# 1. Load the Data
# ==============================================================================
data = lifelines.datasets.load_gbsg2()
# preprocessing
data.rename(columns={'cens': 'event'}, inplace=True)
data['horTh'] = data['horTh'].map({'no': 0, 'yes': 1})
data['menostat'] = data['menostat'].map({'Pre': 0, 'Post': 1})
data['tgrade'] = data['tgrade'].map({'I': 1, 'II': 2, 'III': 3})
# randomly divide the data into training and validation sets
df_train = data.sample(frac=0.7, random_state=42)  # 70% for training
df_train = df_train.reset_index(drop=True)
df_test = data.drop(df_train.index)  # remaining 30% for testing
df_test = df_test.reset_index(drop=True)
x_test = df_test.drop(columns=['time', 'event']).values

# ==============================================================================
# 2. Fit a Cox Proportional Hazards Model
# ==============================================================================
cph = CoxPHFitter()
cph.fit(df_train, duration_col='time', event_col='event')
year = 3  # Set the target year for calibration
# Define time points for 1, 2, 3, 4, 5, 6 years (in days)
times = 365 * year * np.arange(1, 7)

# Predict the survival probability at the given time points
survs_cox = cph.predict_survival_function(x_test)

eval = LifelinesEvaluator(survs_cox, df_test['time'].values, df_test['event'].values)

cox_residuals = eval.residuals(method="CoxSnell", draw_figure=False)
cox2_residuals = eval.residuals(method="Modified CoxSnell-v2", draw_figure=False)
martingale_rs = eval.residuals(method="Martingale", draw_figure=False)
deviance_res = eval.residuals(method="Deviance", draw_figure=False)

cum_haz_empirical = NelsonAalen(cox_residuals, df_test['event'].values)
max_res = np.max(cum_haz_empirical.cumulative_hazard)

fig, ax = plt.subplots(nrows=1, ncols=1, tight_layout=True, dpi=400, figsize=(3.2, 3),)
ax.plot(cum_haz_empirical.survival_times, cum_haz_empirical.cumulative_hazard, color='skyblue')
ax.plot([0, max_res], [0, max_res], label='Ideal', linestyle='dashed', color='grey', linewidth=1)
ax.set_xlim(- max_res * 0.05, max_res * 1.05)
ax.set_ylim(- max_res * 0.05, max_res * 1.05)
ax.legend()
ax.set_xlabel('Cox-Snell Residuals')
ax.set_ylabel('Cumlative hazard of \nCox-Snell residuals')
plt.savefig('cox_snell_residual.png', dpi=400)

idx = np.arange(len(cox_residuals))
event_indicators = df_test['event'].values.astype(bool)


def scatter_plus_hist(
        fig, gs_cell, x_idx, y_vals, events, *,
                      y_label="", title="",
                      bins='auto', alpha=0.5):
    """
    Draw a scatterplot of residuals in the left‑hand panel and, in
    the right‑hand panel, a horizontal histogram of the same y values.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The parent figure to which the axes belong.
    gs_cell : matplotlib.gridspec.SubplotSpec
        A single slot (cell) of a top‑level GridSpec—e.g., outer[i].
    x_idx : 1‑D array‑like
        X‑axis positions (e.g., index or time).
    y_vals : 1‑D array‑like
        Y‑axis data (residuals).
    events : 1‑D boolean array‑like
        True for events, False for censored.
    y_label : str, optional
        Label for the scatter’s y‑axis.
    title : str, optional
        Title for the scatter panel.
    color : str or tuple, optional
        Color for event markers and histogram bars.
    bins : int, str, or sequence, optional
        Passed to ``plt.hist``.
    alpha : float, optional
        Transparency for scatter markers.

    Returns
    -------
    ax_scatter, ax_hist : tuple of matplotlib.axes.Axes
        The two axes objects that were created.
    """
    # Create a 1×2 inner GridSpec within this outer cell
    inner = gs_cell.subgridspec(1, 2, width_ratios=[2, 1], wspace=0.00005)

    # Axes: scatter on the left, histogram on the right
    ax_scatter = fig.add_subplot(inner[0, 0])
    ax_hist    = fig.add_subplot(inner[0, 1], sharey=ax_scatter)

    # Scatter plot
    ax_scatter.scatter(x_idx[events], y_vals[events],  alpha=alpha, facecolors='#D5E8D4', edgecolors='#82B366', linewidths=1, label='Event')
    ax_scatter.scatter(x_idx[~events], y_vals[~events], alpha=alpha, facecolors='#FFF2CC', edgecolors='#D6B656', linewidths=1, label='Censored')
    ax_scatter.set_xlabel('Index')
    ax_scatter.set_ylabel(y_label)
    if title:
        ax_scatter.set_title(title, pad=4, fontsize=9)

    # Histogram (horizontal)
    ax_hist.hist(y_vals, bins=bins, orientation='horizontal', linewidth=1, edgecolor='black', color='#CDE7F0')
    ax_hist.set_xlabel('Count')
    ax_hist.tick_params(axis='y', labelleft=False)  # hide duplicate ticks

    return ax_scatter, ax_hist


fig = plt.figure(figsize=(8, 3), dpi=400, constrained_layout=True)
outer = fig.add_gridspec(1, 3, wspace=0.25)

scatter_plus_hist(fig, outer[0], idx, cox2_residuals, event_indicators,
                  y_label="(a) Modified Cox‑Snell residuals")
scatter_plus_hist(fig, outer[1], idx, martingale_rs, event_indicators,
                  y_label="(b) Martingale residuals")
scatter_plus_hist(fig, outer[2], idx, deviance_res, event_indicators,
                  y_label="(c) Deviance residuals")

# plt.tight_layout()
plt.savefig("combined_residuals.png", dpi=400)



import time
import numpy as np
import warnings
from scipy import integrate

from SurvivalEVAL.Evaluations.util import predict_rmst, predict_mean_st, predict_median_st, predict_median_st, interpolated_survival_curve

def f(x):
    return x*x

def predict_mean_st_old(
        survival_curve: np.ndarray,
        times_coordinate: np.ndarray,
        interpolation: str = "Linear"
) -> float:
    """
    Get the mean survival time from the survival curve. The mean survival time is defined as the area under the survival
    curve. The curve is first interpolated by the given monotonic cubic interpolation method (Linear or Pchip). Then the
    curve gets extroplated by the linear function of (0, 1) and the last time point. The area is calculated by the
    trapezoidal rule.
    Parameters
    ----------
    survival_curve: np.ndarray
        The survival curve of the sample. 1-D array.
    times_coordinate: np.ndarray
        The time coordinate of the survival curve. 1-D array.
    interpolation: str
        The monotonic cubic interpolation method. One of ['Linear', 'Pchip']. Default: 'Linear'.
        If 'Linear', use the interp1d method from scipy.interpolate.
        If 'Pchip', use the PchipInterpolator from scipy.interpolate.
    Returns
    -------
    mean_survival_time: float
        The mean survival time.
    """
    # deprecated warning
    warnings.warn("This function is deprecated. Use 'predict_mean_st' instead.", DeprecationWarning)

    # If all the predicted probabilities are 1 the integral will be infinite.
    if np.all(survival_curve == 1):
        warnings.warn("All the predicted probabilities are 1, the integral will be infinite.")
        return np.inf

    spline = interpolated_survival_curve(times_coordinate, survival_curve, interpolation)

    # predicting boundary
    max_time = float(max(times_coordinate))

    # simply calculate the slope by using the [0, 1] - [max_time, S(t|x)]
    slope = (1 - np.array(spline(max_time)).item()) / (0 - max_time)

    # zero_probability_time = min(times_coordinate[np.where(survival_curve == 0)],
    #                             max_time + (0 - np.array(spline(max_time)).item()) / slope)
    if 0 in survival_curve:
        zero_probability_time = min(times_coordinate[np.where(survival_curve == 0)])
    else:
        zero_probability_time = max_time + (0 - np.array(spline(max_time)).item()) / slope

    def _func_to_integral(time, maximum_time, slope_rate):
        return np.array(spline(time)).item() if time < maximum_time else (1 + time * slope_rate)
    # _func_to_integral = lambda time: spline(time) if time < max_time else (1 + time * slope)
    # limit controls the subdivision intervals used in the adaptive algorithm.
    # Set it to 1000 is consistent with Haider's R code
    mean_survival_time, *rest = integrate.quad(_func_to_integral, 0, zero_probability_time,
                                               args=(max_time, slope), limit=1000)
    return mean_survival_time


if __name__ == '__main__':
    # generate some random points
    logits = np.random.weibull(a=1, size=2000).reshape(100, 20)
    # geenrate pdf using softmax
    pdf = np.exp(logits) / np.exp(logits).sum(axis=1)[:, None]
    surv = 1 - np.cumsum(pdf, axis=1)
    surv = surv[:, :10]

    # generate some random time points
    time_points = np.random.uniform(0, 6, size=1000).reshape(100, 10)
    time_points = np.cumsum(time_points, axis=1)

    # add 1 at the start of the surv and add 0 at the start of time_points
    surv = np.concatenate([np.ones((100, 1)), surv], axis=1)
    time_points = np.concatenate([np.zeros((100, 1)), time_points], axis=1)

    # add a all 1s surv at the end
    surv = np.concatenate([surv, np.ones((1, 11))], axis=0)
    time_points = np.concatenate([time_points, time_points[-1, :].reshape(1, -1)], axis=0)

    start = time.time()
    rmst = predict_rmst(surv, time_points)
    end = time.time()
    print(f"Time taken for RMST: {end - start}")
    rmst_ = np.empty_like(rmst)
    for i in range(surv.shape[0]):
        rmst_[i] = predict_rmst(surv[i], time_points[i])
    assert np.all(rmst == rmst_), "RMST is not correct."

    start = time.time()
    mean_st = predict_mean_st(surv, time_points)
    end = time.time()
    print(f"Time taken for mean survival time: {end - start}")
    strat = time.time()
    mean_st_ = np.empty_like(mean_st)
    for i in range(surv.shape[0]):
        mean_st_[i] = predict_mean_st_old(surv[i], time_points[i])
    end = time.time()
    print(f"Time taken for mean survival time: {end - start}")

    # assert np.all(mean_st == mean_st_), "Mean survival time is not correct."

    start = time.time()
    median_st = predict_median_st(surv, time_points)
    end = time.time()
    median_st_ = np.empty_like(mean_st_)
    for i in range(surv.shape[0]):
        median_st_[i] = predict_median_st(surv[i], time_points[i])
    print(f"Time taken for median survival time: {end - start}")










import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import lifelines
from lifelines import CoxPHFitter

from SurvivalEVAL import LifelinesEvaluator
from SurvivalEVAL.NonparametricEstimator.SingleEvent import  NelsonAalen

color_list = sns.color_palette(n_colors=6)

plt.style.use('seaborn-v0_8-paper')
plt.rcParams["figure.figsize"] = (8, 3.5)
plt.rc('axes', titlesize=9.5)     # fontsize of the axes title
plt.rc('axes', labelsize=9.5)    # fontsize of the x and y labels
# plt.rc('xtick', labelsize=10)    # fontsize of the tick labels
# plt.rc('ytick', labelsize=10)    # fontsize of the tick labels
plt.rc('legend', fontsize=9.5)    # legend fontsize
plt.rc('figure', titlesize=9.5)  # fontsize of the figure title

plt.rcParams['font.size'] = 12


# ==============================================================================
# 1. Load the Data
# ==============================================================================
data = lifelines.datasets.load_gbsg2()
# preprocessing
data.rename(columns={'cens': 'event'}, inplace=True)
data['horTh'] = data['horTh'].map({'no': 0, 'yes': 1})
data['menostat'] = data['menostat'].map({'Pre': 0, 'Post': 1})
data['tgrade'] = data['tgrade'].map({'I': 1, 'II': 2, 'III': 3})
# randomly divide the data into training and validation sets
df_train = data.sample(frac=0.7, random_state=42)  # 70% for training
df_train = df_train.reset_index(drop=True)
df_test = data.drop(df_train.index)  # remaining 30% for testing
df_test = df_test.reset_index(drop=True)
x_test = df_test.drop(columns=['time', 'event']).values

# ==============================================================================
# 2. Fit a Cox Proportional Hazards Model
# ==============================================================================
cph = CoxPHFitter()
cph.fit(df_train, duration_col='time', event_col='event')
year = 3  # Set the target year for calibration
# Define time points for 1, 2, 3, 4, 5, 6 years (in days)
times = 365 * year * np.arange(1, 7)

# Predict the survival probability at the given time points
survs_cox = cph.predict_survival_function(x_test)

eval = LifelinesEvaluator(survs_cox, df_test['time'].values, df_test['event'].values,
                          predict_time_method='Median')

p, stats = eval.log_rank()
print(f'Log-rank test p-value: {p}, statistic: {stats}')
p, stats = eval.log_rank(weightings='wilcoxon')
print(f'Log-rank test p-value: {p}, statistic: {stats}')
p, stats = eval.log_rank(weightings='tarone-ware')
print(f'Log-rank test p-value: {p}, statistic: {stats}')
p, stats = eval.log_rank(weightings='peto')
print(f'Log-rank test p-value: {p}, statistic: {stats}')
p, stats = eval.log_rank(weightings='fleming-harrington', p=0.1, q=2)
print(f'Log-rank test p-value: {p}, statistic: {stats}')

pred_times = eval.predicted_event_times
cum_haz_empirical = NelsonAalen(df_test['time'].values, df_test['event'].values)
cum_haz_pred = NelsonAalen(pred_times, np.ones_like(df_test['event'].values, dtype=bool))
max_x = np.max(df_test['time'].values)
max_y = np.max(cum_haz_empirical.cumulative_hazard)

fig, ax = plt.subplots(nrows=1, ncols=1, tight_layout=True, dpi=400, figsize=(3.2, 3),)
ax.plot(cum_haz_empirical.survival_times, cum_haz_empirical.cumulative_hazard, color='grey', label='ideal (NA)', linewidth=2, linestyle='dashed')
ax.plot(cum_haz_pred.survival_times, cum_haz_pred.cumulative_hazard, color='skyblue', label='predicted', linewidth=2)
ax.legend()
ax.set_xlim(- max_x * 0.05, max_x * 1.05)
ax.set_ylim(- max_y * 0.05, max_y * 1.05)
ax.set_xlabel('Time (days)')
ax.set_ylabel('Cumlative hazard')
plt.savefig('log_rank_test.png', dpi=400)




from typing import Optional
import numpy as np
import pandas as pd
from lifelines import WeibullAFTFitter
from SurvivalEVAL.Evaluations.custom_types import Numeric
from SurvivalEVAL.NonparametricEstimator.SingleEvent import KaplanMeier, TurnbullEstimator


def single_brier_score(
        preds: np.ndarray,
        event_times: np.ndarray,
        event_indicators: np.ndarray,
        train_event_times: np.ndarray,
        train_event_indicators: np.ndarray,
        target_time: Optional[float] = None,
        ipcw: bool = True
) -> float:
    """
    Calculate the Brier score at a specific time.

    Parameters
    ----------
    preds: np.ndarray, shape = (n_samples, )
        Estimated survival probabilities at the specific time for the testing samples.
    event_times: np.ndarray, shape = (n_samples, )
        Actual event/censor time for the testing samples.
    event_indicators: np.ndarray, shape = (n_samples, )
        Binary indicators of censoring for the testing samples
    train_event_times: np.ndarray, shape = (n_train_samples, )
        Actual event/censor time for the training samples.
    train_event_indicators: np.ndarray, shape = (n_train_samples, )
        Binary indicators of censoring for the training samples
    target_time: float, default: None
        The specific time point for which to estimate the Brier score.
    ipcw: bool, default: True
        Whether to use Inverse Probability of Censoring Weighting (IPCW) in the calculation.

    Returns
    -------
    brier_score: float
        Value of the brier score.
    """
    if target_time is None:
        target_time = np.median(event_times)

    event_indicators = event_indicators.astype(bool)
    # train_event_indicators = train_event_indicators.astype(bool)

    if ipcw:
        inverse_train_event_indicators = 1 - train_event_indicators
        ipc_model = KaplanMeier(train_event_times, inverse_train_event_indicators)

        ipc_pred = ipc_model.predict(event_times)
        # Catch if denominator is 0.
        ipc_pred[ipc_pred == 0] = np.inf
        # Category one calculates IPCW weight at observed time point.
        # Category one is individuals with event time lower than the time of interest and were NOT censored.
        weight_cat1 = ((event_times <= target_time) & event_indicators) / ipc_pred
        # Catch if event times goes over max training event time, i.e. predict gives NA
        weight_cat1[np.isnan(weight_cat1)] = 0
        # Category 2 is individuals whose time was greater than the time of interest (singleBrierTime)
        # contain both censored and uncensored individuals.
        weight_cat2 = (event_times > target_time) / ipc_model.predict(target_time)
        # predict returns NA if the passed-in time is greater than any of the times used to build the inverse probability
        # of censoring model.
        weight_cat2[np.isnan(weight_cat2)] = 0
    else:
        weight_cat1 = ((event_times <= target_time) & event_indicators)
        weight_cat2 = (event_times > target_time)

    b_score = (np.square(preds) * weight_cat1 + np.square(1 - preds) * weight_cat2).mean()
    ###########################
    # Here we are ordering event times and then using predict with level.chaos = 1 which returns
    # predictions ordered by time.
    # This is from Haider's code in R, but I feel it doesn't need to be ordered by time.
    # Refer above few lines for the justified code
    ###########################
    # order_of_times = np.argsort(event_times)
    # # Catch if event times goes over max training event time, i.e. predict gives NA
    # weight_cat1 = ((event_times[order_of_times] <= target_time) & event_indicators[order_of_times]) /\
    #               ipc_model.predict(event_times[order_of_times])
    # weight_cat1[np.isnan(weight_cat1)] = 0
    # weight_cat2 = (event_times[order_of_times] > target_time) / ipc_model.predict(target_time)
    # weight_cat2[np.isnan(weight_cat2)] = 0
    #
    # survival_curves_ordered = survival_curves[order_of_times, :]
    # predict_probs = []
    # for i in range(survival_curves_ordered.shape[0]):
    #     predict_prob = predict_prob_from_curve(survival_curves_ordered[i, :], time_coordinates,
    #                                            event_times[order_of_times][i])
    #     predict_probs.append(predict_prob)
    # predict_probs = np.array(predict_probs)
    #
    # b_score = np.mean(np.square(predict_probs) * weight_cat1 + np.square(1 - predict_probs) * weight_cat2)
    return b_score


def brier_score_ic(
        preds: np.ndarray,
        left_limits: np.ndarray,
        right_limits: np.ndarray,
        train_left_limits: Optional[np.ndarray] = None,
        train_right_limits: Optional[np.ndarray] = None,
        x: Optional[np.ndarray] = None,
        x_train: Optional[np.ndarray] = None,
        target_time: Optional[Numeric] = None,
        method: str = "Tsouprou-marginal",
) -> float:
    """
    Calculate the Brier score at a specific time.

    Parameters
    ----------
    preds: np.ndarray, shape = (n_samples, )
        Estimated survival probabilities at the specific time for the testing samples.
    left_limits: np.ndarray, shape = (n_samples, )
        Actual left limit event/censor time for the testing samples.
    right_limits: np.ndarray, shape = (n_samples, )
        Actual right limit event/censor time for the testing samples.
    train_left_limits: np.ndarray, shape = (n_train_samples, )
        Actual left limit event/censor time for the training samples.
    train_right_limits: np.ndarray, shape = (n_train_samples, )
        Actual right limit event/censor time for the training samples.
    x: np.ndarray, shape = (n_samples, n_features), default: None
        Features for the testing samples. Use only when method is 'Tsouprou-conditional'.
    x_train: np.ndarray, shape = (n_train_samples, n_features), default: None
        Features for the training samples. Use only when method is 'Tsouprou-conditional'.
    target_time: numeric, default: None
        The specific time point for which to estimate the Brier score.
    method: str, default: IPCW
        Method to use for handling censoring. One of ['uncensored', 'Tsouprou-marginal', 'Tsouprou-conditional'].
        'uncensored': Treat censored data as uncensored.
        'Tsouprou-marginal': Use marginal survival probabilities based on Turnbull estimator.
        'Tsouprou-conditional': Use conditional survival probabilities based on Weibull AFT model.
    Returns
    -------
    brier_score: float
        Value of the brier score.
    """
    if target_time is None:
        tau_vals = np.concatenate([left_limits, right_limits[np.isfinite(right_limits)]])
        tau = np.unique(np.sort(tau_vals))
        target_time = np.median(tau)

    if method == "uncensored":
        # if the target time is within the interval, then we calculate the brier score
        # otherwise, we skip the instance
        mask = (left_limits <= target_time) & (right_limits > target_time)
        weight = 1 - mask.astype(float)
        # get the survival status at the target time
        # if the left limit is greater than the target time, then the event has not occurred, so 1
        # if the right limit is less than or equal to the target time, then the event has occurred, so 0
        survival_status = (left_limits > target_time).astype(float)
        brier_score = (np.square(preds - survival_status) * weight).sum() / weight.sum()
    elif "Tsouprou" in method:
        # method based on Sofia Tsouprou's thesis
        # Measures of discrimination and predictive accuracy for interval censored survival data
        # https://studenttheses.universiteitleiden.nl/access/item:3597164/view
        # the original method uses Weibull parametric model to estimate the survival function
        # here we give the option to use either Turnbull estimator or Weibull AFT model

        # must have training data
        if train_left_limits is None or train_right_limits is None:
            raise ValueError("Training data must be provided for Tsouprou methods.")

        if method == "Tsouprou-marginal":
            marginal_estimator = TurnbullEstimator().fit(
                left=train_left_limits,
                right=train_right_limits,
            )
            # get the marginal survival probabilities at the target time
            left_probs = marginal_estimator.predict(left_limits)
            right_probs = marginal_estimator.predict(right_limits)
            target_probs = marginal_estimator.predict(target_time)
        elif method == "Tsouprou-conditional":
            if x is None or x_train is None:
                raise ValueError("Features for both training and testing data must be provided for "
                                 "Tsouprou-conditional method.")

            if x.ndim != x_train.ndim:
                raise ValueError("x and x_train must have the same number of dimensions.")

            train_data = {
                'left': train_left_limits,
                'right': train_right_limits
            }
            if x_train.ndim == 1:
                train_data['feature'] = x_train
            elif x_train.ndim == 2:
                for i in range(x_train.shape[1]):
                    train_data[f'feature_{i}'] = x_train[:, i]
            else:
                raise ValueError("x_train must be a 1-D or 2-D array.")
            train_df = pd.DataFrame(train_data)

            x_data = {}
            if x.ndim == 1:
                x_data['feature'] = x
            elif x.ndim == 2:
                for i in range(x.shape[1]):
                    x_data[f'feature_{i}'] = x[:, i]
            else:
                raise ValueError("x must be a 1-D or 2-D array.")
            x_df = pd.DataFrame(x_data)

            aft_model = WeibullAFTFitter()
            aft_model.fit_interval_censoring(train_df, 'left', 'right')
            # get the conditional survival probabilities at the left limit, right limit, and target time
            left_probs = aft_model.predict_survival_function(x_df, times=left_limits).values.diagonal()
            right_probs = aft_model.predict_survival_function(x_df, times=right_limits).values.diagonal()
            target_probs = aft_model.predict_survival_function(x_df, target_time).values.flatten()
        else:
            raise ValueError(f"Method {method} is not supported.")
        # exam on non-bad indices
        # bad indices are those (1) the target time is within the interval and (2) the left and right survival
        # probabilities are the same, which leads to zeros in both numerator and denominator in survival_status
        bad = (left_probs == right_probs) & (left_limits < target_time) & (target_time <= right_limits)
        if np.sum(bad) > 0:
            left_limits = left_limits[~bad]
            right_limits = right_limits[~bad]
            preds = preds[~bad]
            left_probs = left_probs[~bad]
            right_probs = right_probs[~bad]
            if isinstance(target_time, np.ndarray):
                target_probs = target_probs[~bad]

        # supress warnings for divide by zero
        with np.errstate(divide='ignore', invalid='ignore'):
            survival_status = (target_probs - right_probs) / (left_probs - right_probs)
        survival_status[right_limits < target_time] = 0
        survival_status[left_limits >= target_time] = 1

        if np.any((survival_status < 0) | (survival_status > 1)):
            raise ValueError("Calculated survival status contains values outside [0, 1].")

        # calculate the brier score
        brier_score = np.mean(np.square(preds - survival_status))
    else:
        raise ValueError(f"Method {method} is not supported.")
    return brier_score


def brier_multiple_points(
        pred_mat: np.ndarray,
        event_times: np.ndarray,
        event_indicators: np.ndarray,
        train_event_times: Optional[np.ndarray],
        train_event_indicators: Optional[np.ndarray],
        target_times: np.ndarray,
        ipcw: bool = True
) -> np.ndarray:
    """
    Calculate multiple Brier scores at multiple specific times.

    Parameters
    ----------
    pred_mat: np.ndarray, shape = (n_samples, n_time_points)
        Predicted probability array (2-D) for each instances at each time point.
    event_times: np.ndarray, shape = (n_samples, )
        Actual event/censor time for the testing samples.
    event_indicators: np.ndarray, shape = (n_samples, )
        Binary indicators of censoring for the testing samples
    train_event_times: np.ndarray, shape = (n_train_samples, )
        Actual event/censor time for the training samples.
    train_event_indicators: np.ndarray, shape = (n_train_samples, )
        Binary indicators of censoring for the training samples
    target_times: float
        The specific time points for which to estimate the Brier scores.
    ipcw: bool, default: True
        Whether to use Inverse Probability of Censoring Weighting (IPCW) in the calculation.

    Returns
    -------
    brier_scores: np.ndarray, shape = (n_time_points, )
        Values of multiple Brier scores.
    """
    if target_times.ndim != 1:
        error = "'time_grids' is not a one-dimensional array."
        raise TypeError(error)

    # bs_points_matrix = np.tile(event_times, (len(target_times), 1))
    target_times_mat = np.repeat(target_times.reshape(1, -1), repeats=len(event_times), axis=0)
    event_times_mat = np.repeat(event_times.reshape(-1, 1), repeats=len(target_times), axis=1)
    event_indicators_mat = np.repeat(event_indicators.reshape(-1, 1), repeats=len(target_times), axis=1)
    event_indicators_mat = event_indicators_mat.astype(bool)

    if ipcw:
        if train_event_times is None or train_event_indicators is None:
            raise ValueError("Training event times and indicators must be provided for IPCW weighting.")

        inverse_train_event_indicators = 1 - train_event_indicators

        ipc_model = KaplanMeier(train_event_times, inverse_train_event_indicators)

        # Category one calculates IPCW weight at observed time point.
        # Category one is individuals with event time lower than the time of interest and were NOT censored.
        ipc_pred = ipc_model.predict(event_times_mat)
        # Catch if denominator is 0.
        ipc_pred[ipc_pred == 0] = np.inf
        weight_cat1 = ((event_times_mat <= target_times_mat) & event_indicators_mat) / ipc_pred
        # Catch if event times goes over max training event time, i.e. predict gives NA
        weight_cat1[np.isnan(weight_cat1)] = 0
        # Category 2 is individuals whose time was greater than the time of interest (singleBrierTime)
        # contain both censored and uncensored individuals.
        ipc_target_pred = ipc_model.predict(target_times_mat)
        # Catch if denominator is 0.
        ipc_target_pred[ipc_target_pred == 0] = np.inf
        weight_cat2 = (event_times_mat > target_times_mat) / ipc_target_pred
        # predict returns NA if the passed in time is greater than any of the times used to build
        # the inverse probability of censoring model.
        weight_cat2[np.isnan(weight_cat2)] = 0
    else:
        weight_cat1 = ((event_times_mat <= target_times_mat) & event_indicators_mat)
        weight_cat2 = (event_times_mat > target_times_mat)

    ipcw_square_error_mat = np.square(pred_mat) * weight_cat1 + np.square(1 - pred_mat) * weight_cat2
    brier_scores = np.mean(ipcw_square_error_mat, axis=0)
    return brier_scores



import numpy as np
from SurvivalEVAL import PointEvaluator

np.random.seed(1)
# get some synthetic data
n_data = 1000
n_features = 10
event_times = np.random.uniform(0, 6, size=n_data).round(1)
censoring_times = np.random.lognormal(mean=1, sigma=1, size=n_data).round(1)
event_indicators = event_times < censoring_times
observed_times = np.minimum(event_times, censoring_times)

# get training data
n_train = 800
train_event_times = event_times[:n_train]
train_event_indicators = event_indicators[:n_train]

# get testing data
test_event_times = event_times[n_train:]
test_event_indicators = event_indicators[n_train:]

# get some synthetic predictions
predicted_times = np.random.uniform(0, 6, size=200).round(1)

# initialize the evaluator
evaluator = PointEvaluator(predicted_times, test_event_times, test_event_indicators,
                           train_event_times, train_event_indicators)
print("Successfully initialized the evaluator.")

# calculate the concordance index
cindex, _, _ = evaluator.concordance(method="Margin")
print(f"The concordance index is {cindex}.")

# calculate the MAE
mae = evaluator.mae(method="Pseudo_obs", weighted=True)
print(f"The MAE is {mae}.")

# calculate the rmse
rmse = evaluator.rmse(method="Margin", weighted=False)
print(f"The RMSE is {rmse}.")

# calculate the log-rank test p-value
p_value, _ = evaluator.log_rank()
print(f"The log-rank test p-value is {p_value}.")

# plot the figure for kaplan-meier curves for true labels and predicted times
import matplotlib.pyplot as plt
from lifelines import KaplanMeierFitter
km1 = KaplanMeierFitter().fit(test_event_times, test_event_indicators)
km2 = KaplanMeierFitter().fit(predicted_times, np.ones_like(predicted_times))
fig, ax= plt.subplots()
km1.plot_survival_function(ax=ax, label="True")
km2.plot_survival_function(ax=ax, label="Predicted")
plt.title("Kaplan-Meier Curves")
plt.xlabel("Time")
plt.ylabel("Survival Probability")
plt.legend()
plt.show()



